/*
 * @(#)DrawApplication.java 5.1
 *
 */

package CH.ifa.draw.application;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.contrib.*;
import CH.ifa.draw.util.*;

// for class DrawingLoadServer:
import java.io.*;
import java.net.*;
import java.util.*;

import de.renew.util.StringUtil;

/**
 * DrawApplication defines a standard presentation for
 * standalone drawing editors. The presentation is
 * customized in subclasses.
 * The application is started as follows:
 * <pre>
 * public static void main(String[] args) {
 *     MyDrawApp window = new MyDrawApp();
 * }
 * </pre>
 */

public  class DrawApplication
        implements DrawingEditor, PaletteListener {

    private static int windowCount = 0;

    protected Frame             menuFrame, toolFrame;
    private final String        defaultDrawExtension="draw";
    public String               drawExtension;
    private Font                fMenuFont;
    private boolean             floating;
    private Vector              fDrawings = new Vector();
    private Tool                fTool;
    private Iconkit             fIconkit;
    private Menu                fDrawingsMenu;
    private TextField           fStatusLine;
    private Vector              fViewFrames = new Vector();
    private DrawingViewFrame    fViewFrame;
    private boolean             fToolButtonSticky=false;
    private ToolButton          fDefaultToolButton;
    private ToolButton          fSelectedToolButton = null;
    private ToolButton          fConnTextTB, fTextTB;

    // the image resource path
    private static final String fgDrawPath = "/CH/ifa/draw/";
    public static final String IMAGES = fgDrawPath+"images/";

    /**
     * The index of the file menu in the menu bar.
     */
    public static final int    FILE_MENU = 0;
    /**
     * The index of the edit menu in the menu bar.
     */
    public static final int    EDIT_MENU = 1;
    /**
     * The index of the alignment menu in the menu bar.
     */
    public static final int    ALIGNMENT_MENU = 2;
    /**
     * The index of the attributes menu in the menu bar.
     */
    public static final int    ATTRIBUTES_MENU = 3;

    /**
     * The index of the drawings menu in the menu bar.
     */
    public static final int    DRAWINGS_MENU = 5;

    private String fLastPath = System.getProperty("user.dir");

    /** Lookup table for default directories keyed by file extensions. **/
    private Hashtable dirByExt = new Hashtable();

    /** Default directory used if the lookup table has no entry for some extension. **/
    private String fallbackDir = null;

    /**
     * Constructs a drawing window with a default title.
     */
    public DrawApplication() {
       this("JHotDraw");
    }

    /**
     * Constructs a drawing window with the given title.
     */
    public DrawApplication(String title) {
       this(title, null, new String[0], null);
    }

    public DrawApplication(String title, String[] drawings) {
       this(title,null,drawings,null);
    }

  protected Component wrapStatusLine(Component statusLine) {
    return statusLine;
  }

    public DrawApplication(String title, String drawExtension, String[] drawings, String icon) {

        DrawingLoadServer dls=null;
        String serverPortStr=System.getProperty("de.renew.loadServerPort");
        if (serverPortStr!=null) {
          try {
             int serverPort=Integer.valueOf(serverPortStr).intValue();
             try {
                dls=new DrawingLoadServer(this,serverPort);
             } catch (IOException ioe) {
                System.out.println("Server Socket is occupied!");
                System.exit(-1);
             }
          } catch (Exception e) {
             System.out.println("No Drawing Load Server: "+e);
          }
        } /* endif */

        menuFrame = new Frame(title);
        fStatusLine = createStatusLine();
        menuFrame.add("South", wrapStatusLine(fStatusLine));

//        toolFrame = new Frame("Tools");
//        toolFrame.setLayout(new BorderLayout());
        toolFrame = menuFrame;
        floating = toolFrame == menuFrame;
        fIconkit = new Iconkit(toolFrame);
        if (icon != null) {
            Image iconImage = fIconkit.loadImageResource(icon);
            if (iconImage == null) {
               System.err.println("Resource icon "+icon+" could not be loaded!");
            } else {
               menuFrame.setIconImage(iconImage);
            }
        }
        Panel tools = createToolPalette();
        createTools(tools);
        toolFrame.add("Center", tools);

	MenuBar mb = new MenuBar();
	String fontSizeStr=System.getProperty("de.renew.menuFontSize");
        fMenuFont=mb.getFont();
	if (fontSizeStr!=null) {
	    int fontSize;
	    try {
		fontSize=Integer.parseInt(fontSizeStr);
		System.out.println("Setting menu font size to "+fontSize+" pt.");
		fMenuFont=new Font("SansSerif",Font.PLAIN,fontSize);
	    } catch (NumberFormatException e) {
		System.out.println("Failed to set menu font size to \""+fontSizeStr+"\": wrong number format.");
	    }
	}
	createMenus(mb);
	menuFrame.setMenuBar(mb);

        menuFrame.pack();

	// Workaround for Windows JDK pack()-bug:
	// The menu-frame is too small or too large.
        String windowResizable=System.getProperty("de.renew.windowResizable");
	//if (windowResizable==null) {
            // If no parameter is given, the window is
            // resizable on Windows-Systems only.
	//if (System.getProperty("os.name").indexOf("Win")>=0) {
	//       windowResizable="true";
        //    }
	//}
        menuFrame.setResizable(Boolean.valueOf(windowResizable).booleanValue());

        //Dimension d = defaultSize();
	//setSize(d.width, d.height);
        menuFrame.show();
	menuFrame.pack(); // has to be done after show() on Windows-Systems

	// This should show the main window. However, there were some problems
	// in the very early versions of Renew, where the appearance of the
	// main window would depend crucially on the order of some
	// commands. It is not clear whether the errors can occur again.
	// This must be monitored.

        if (toolFrame != menuFrame) {
           toolFrame.pack();
           toolFrame.show();
        } /* endif */

	windowCount++;
        addListeners();

        if (drawExtension==null)
            this.drawExtension=defaultDrawExtension;
        else    
           this.drawExtension=drawExtension;

	//System.out.println("number of params: "+drawings.length);
        Drawing drawing;
        for (int i=0; i<drawings.length; ++i)
           loadAndOpenDrawing(drawings[i]);
        if (fDrawings.size() == 0)
           promptNew();

        toolDone();

	if (dls!=null) {
	  dls.start();
	}
    }

    /**
     * Registers the listeners for the main windows
     */
    protected void addListeners() {
	    menuFrame.addWindowListener(
              new WindowAdapter() {
                public void windowClosing(WindowEvent event) {
                    exit();
                }
              }
            );
            if (toolFrame != menuFrame) {
	      toolFrame.addWindowListener(
                new WindowAdapter() {
                  public void windowClosing(WindowEvent event) {
		    toolFrame.setVisible(false);
                 }
                }
              );
            } /* endif */
    }

    public Frame menuFrame() { return menuFrame; }

    protected DrawApplication newWindow() { return new DrawApplication(menuFrame.getTitle()); }

//     private Menu titleSubstitute;

//     public synchronized void setTitle(String title) {
//       menuFrame.setTitle(title);
//       titleSubstitute.setLabel(title);
//     }

    private void updateName(Drawing drawing, String filename) {
      showDrawingViewFrame(drawing);
      filename=StringUtil.makeCanonical(filename);
      fLastPath = StringUtil.getPath(filename);
      String nameOnly = StringUtil.getFilename(filename);
      fViewFrame.setTitle(nameOnly);
      drawing.setName(nameOnly);
      drawing.setFilename(filename);
      int pos = fDrawings.indexOf(drawing);
      fDrawingsMenu.getItem(pos).setLabel(nameOnly);
    }

    /**
     * Creates the standard menus. Clients override this
     * method to add additional menus.
     */
    protected void createMenus(MenuBar mb) {
		mb.add(createFileMenu());
		mb.add(createEditMenu());
		mb.add(createAlignmentMenu());
		mb.add(createAttributesMenu());
		mb.add(createDebugMenu());
		mb.add(createDrawingsMenu());
    }

    void addDrawing(Drawing drawing) {
       if (!fDrawings.contains(drawing)) {
           fDrawings.addElement(drawing);
           CheckboxMenuItem mi = new CheckboxMenuItem(drawing.getName());
	   mi.setFont(fMenuFont);
           mi.addItemListener(new DrawingMenuListener(this,drawing));
           fDrawingsMenu.add(mi);

	   drawingAdded(drawing);
       } /* endif */
    }

    protected boolean closeViewFrame(DrawingViewFrame viewFrame) {
      Drawing drawing=viewFrame.view().drawing();

      if (drawing.isModified()) {
          int answer = MessageDialog.invokeDialog(
                menuFrame, "Renew: Confirm Close",
                new String[] {"The drawing \""+drawing.getName()+"\"","you are about to close","has been modified."},
                MessageDialog.LEFT, new String[] {" Save now "," Close ", " Cancel "}
          );
          if (answer == 0) {
             if (!saveDrawing(drawing))
                return false;
          } else if (answer == 2) {
            return false;
          }
      }
      viewFrame.setVisible(false);
      viewFrame.dispose();
      fViewFrames.removeElement(viewFrame);
      if (viewFrame == fViewFrame) {
	int numViews=fViewFrames.size();
        if (numViews > 0) {
	  setCurrentDrawing((DrawingViewFrame)fViewFrames.elementAt(numViews-1));
          fViewFrame.show();
	  fViewFrame.requestFocus(); // !!!
	} else
	  setCurrentDrawing(null);
      }

      int pos = fDrawings.indexOf(drawing);
      fDrawingsMenu.remove(pos);
      fDrawings.removeElementAt(pos);

      drawing.release();

      drawingReleased(drawing);

      viewFrame.discard();

      return true;
    }

    protected void drawingAdded(Drawing drawing) {
      // My subclasses will add special behaviour here.
    }

    protected void drawingReleased(Drawing drawing) {
      // My subclasses will add special behaviour here.
    }

    protected Menu createMenu(String name) {
	Menu menu=new Menu(name);
	menu.setFont(fMenuFont);
	return menu;
    }

    public MenuItem createMenuItem(String name, MenuShortcut shortcut,
				      ActionListener action) {
	MenuItem mi;
	if (shortcut==null)
	    mi=new MenuItem(name);
	else
	    mi=new MenuItem(name,shortcut);
	mi.setFont(fMenuFont);
	mi.addActionListener(action);
	return mi;
    }

    public MenuItem createMenuItem(String name, ActionListener action) {
	return createMenuItem(name,null,action);
    }

    public CommandMenu createCommandMenu(String name) {
	CommandMenu cm=new CommandMenu(name);
	cm.setFont(fMenuFont);
	return cm;
    }

    protected Menu createDrawingsMenu() {
        fDrawingsMenu = createMenu("Drawings");
        return fDrawingsMenu;
    }

    protected Enumeration drawings() {
      return fDrawings.elements();
    }

    protected DrawingViewFrame newDrawingViewFrame(Drawing drawing) {
        Dimension d = getDrawingViewSize();
        StandardDrawingView drawingView = createDrawingView(d.width, d.height);
        addDrawing(drawing);
        DrawingViewFrame viewFrame=new DrawingViewFrame(this, drawingView, drawing);
        fViewFrames.addElement(viewFrame);
        setCurrentDrawing(viewFrame);
        return viewFrame;
    }
	  
    /**
     * Creates the file menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createFileMenu() {
		Menu menu = createMenu("File");

		menu.add(createMenuItem("New Drawing", new MenuShortcut('N'),
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            promptNew();
		        }
		    }
					));

// 		menu.add(createMenuItem("New Frame", new MenuShortcut('F'),
// 		    new ActionListener() {
// 		        public void actionPerformed(ActionEvent event) {
// 			  newDrawingViewFrame(drawing());
// 		        }
// 		    }
// 		        ));

		menu.add(createMenuItem("Open Drawing...", new MenuShortcut('O'),
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            promptOpen();
		        }
		    }
					));

		menu.add(createMenuItem("Save Drawing", new MenuShortcut('S'),
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            saveDrawing(drawing());
		        }
		    }
					));

		menu.add(createMenuItem("Save Drawing As...",
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
                            Drawing drawing=drawing();
		            if (promptSaveAs(drawing))
                                saveDrawing(drawing);
		        }
		    }
					));

		menu.add(createMenuItem("Save All Drawings",
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
                            Enumeration drawings=drawings();
                            while (drawings.hasMoreElements()) {
                               Drawing drawing=(Drawing)drawings.nextElement();
                               if (drawing.getFilename()==null) {
                                  showDrawingViewFrame(drawing);
                                  if (!promptSaveAs(drawing))
                                      return;
                               }
                             }
                             drawings=drawings();
                             while (drawings.hasMoreElements()) {
		                saveDrawing((Drawing)drawings.nextElement());
                            }
		        }
		    }
					));

		menu.add(createMenuItem("Close Drawing",
        	     new ActionListener() {
		       public void actionPerformed(ActionEvent event) {
			   if (fViewFrame != null) {
			       closeViewFrame(fViewFrame);
			   }
		       }
		}
					));

//     	menu.add(createMenuItem("Close Frame",
// 	   new ActionListener() {
// 	      public void actionPerformed(ActionEvent event) {
// 		fViewFrame.hide();
//                 fViewFrame.dispose();
// 	      }
// 	   }
//              ));

		menu.addSeparator();

		menu.add(createMenuItem("Print...", new MenuShortcut('P'),
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            print();
		        }
		    }
					));

		menu.addSeparator();

		menu.add(createMenuItem("Exit",
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            exit();
		        }
		    }
					));

		return menu;
	}

    /**
     * Creates the edit menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createEditMenu() {
		CommandMenu menu = createCommandMenu("Edit");
		menu.add(new CutCommand("Cut", this), new MenuShortcut('X'));
		menu.add(new CopyCommand("Copy", this), new MenuShortcut('C'));
		menu.add(new PasteCommand("Paste", this), new MenuShortcut('V'));
		menu.addSeparator();
		menu.add(new DuplicateCommand("Duplicate", this), new MenuShortcut('D'));
		menu.add(new DeleteCommand("Delete", this));
		menu.addSeparator();
		menu.add(new GroupCommand("Group", this));
		menu.add(new UngroupCommand("Ungroup", this));
		menu.addSeparator();
		// The menues "Send To Back" and "Bring To Front" were
		// moved to the "Layout" (formerly "Align") menu.
                menu.add(new SelectCommand("Select All", this), new MenuShortcut('A'));
		menu.addSeparator();

		menu.add(new ToggleStickyCommand("Toggle Sticky Tools",
		  this));
		return menu;
	}

    /**
     * Creates the alignment menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createAlignmentMenu() {
	CommandMenu menu = createCommandMenu("Layout"); // "Align"
	menu.add(new ToggleGridCommand("Toggle Snap to Grid",
				       this, new Point(5,5)));
	menu.add(new SnapToGridCommand("Snap to Grid now",
				       this, new Point(5,5)));
	menu.addSeparator();
	CommandMenu alignMenu = createCommandMenu("Align");
	alignMenu.add(new AlignCommand("Lefts", this, AlignCommand.LEFTS));
	alignMenu.add(new AlignCommand("Centers", this, AlignCommand.CENTERS));
	alignMenu.add(new AlignCommand("Rights", this, AlignCommand.RIGHTS));
	alignMenu.addSeparator();
	alignMenu.add(new AlignCommand("Tops", this, AlignCommand.TOPS));
	alignMenu.add(new AlignCommand("Middles", this, AlignCommand.MIDDLES));
	alignMenu.add(new AlignCommand("Bottoms", this, AlignCommand.BOTTOMS));
        menu.add(alignMenu);
	CommandMenu spreadMenu = createCommandMenu("Spread");
	spreadMenu.add(new SpreadCommand("Lefts", this, SpreadCommand.LEFTS));
	spreadMenu.add(new SpreadCommand("Centers", this, SpreadCommand.CENTERS));
	spreadMenu.add(new SpreadCommand("Rights", this, SpreadCommand.RIGHTS));
	spreadMenu.add(new SpreadCommand("Distances", this, SpreadCommand.HORIZONTAL_DISTANCE));
	spreadMenu.addSeparator();
	spreadMenu.add(new SpreadCommand("Tops", this, SpreadCommand.TOPS));
	spreadMenu.add(new SpreadCommand("Middles", this, SpreadCommand.MIDDLES));
	spreadMenu.add(new SpreadCommand("Bottoms", this, SpreadCommand.BOTTOMS));
	spreadMenu.add(new SpreadCommand("Distances", this, SpreadCommand.VERTICAL_DISTANCE));

	spreadMenu.addSeparator();
	spreadMenu.add(new SpreadCommand("Diagonal", this, SpreadCommand.DIAGONAL_CENTERS));
	//spreadMenu.add(new SpreadCommand("Distances", this, SpreadCommand.DIAGONAL_DISTANCE));
	
        menu.add(spreadMenu);
	// the following menus used to be in "Edit":
	menu.addSeparator();
	menu.add(new SendToBackCommand("Send to Back", this));
	menu.add(new BringToFrontCommand("Bring to Front", this));
	return menu;
    }

    /**
     * Creates the debug menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createDebugMenu() {
		Menu menu = createMenu("Debug");

		menu.add(createMenuItem("Simple Update",
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            view().setDisplayUpdate(new SimpleUpdateStrategy());
		        }
		    }
					));

		menu.add(createMenuItem("Buffered Update",
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            view().setDisplayUpdate(new BufferedUpdateStrategy());
		        }
		    }
					));
		return menu;
	}

    /**
     * Creates the attributes menu and its submenus. Clients override this
     * method to add additional menu items.
     */
    protected Menu createAttributesMenu() {
        Menu menu = createCommandMenu("Attributes");
        menu.add(createColorMenu("Fill Color", "FillColor"));
        menu.add(createColorMenu("Pen Color", "FrameColor"));
        menu.add(createVisibilityMenu());
        menu.add(createArrowMenu());

	menu.addSeparator();

        menu.add(createFontMenu());
        menu.add(createFontSizeMenu());
        menu.add(createFontStyleMenu());
        menu.add(createColorMenu("Text Color", "TextColor"));
        return menu;
    }

    /**
     * Creates the color menu.
     */
    protected Menu createColorMenu(String title, String attribute) {
        CommandMenu menu = createCommandMenu(title);
        for (int i=0; i<ColorMap.size(); i++)
            menu.add(
                new ChangeAttributeCommand(
                    ColorMap.name(i),
                    attribute,
                    ColorMap.color(i),
                    this
                )
            );
        return menu;
    }

    /**
     * Creates the visibility menu.
     */
    protected Menu createVisibilityMenu() {
        CommandMenu menu = createCommandMenu("Visibility");
        menu.add(new ChangeAttributeCommand("visible", "Visibility",
					    Boolean.TRUE, this));
        menu.add(new ChangeAttributeCommand("invisible", "Visibility",
					    Boolean.FALSE, this));
	return menu;
    }

    /**
     * Creates the arrows menu.
     */
    protected Menu createArrowMenu() {
        CommandMenu menu = createCommandMenu("Arrow");
        menu.add(new ChangeAttributeCommand("none",     "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_NONE),  this));
        menu.add(new ChangeAttributeCommand("at Start", "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_START), this));
        menu.add(new ChangeAttributeCommand("at End",   "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_END),   this));
        menu.add(new ChangeAttributeCommand("at Both",  "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_BOTH),  this));
        return menu;
    }

    /**
     * Creates the fonts menus. It installs all available fonts
     * supported by the toolkit implementation.
     */
    protected Menu createFontMenu() {
        CommandMenu menu = createCommandMenu("Font");
        String fonts[] = Toolkit.getDefaultToolkit().getFontList();
        for (int i = 0; i < fonts.length; i++)
            menu.add(new ChangeAttributeCommand(fonts[i], "FontName", fonts[i],  this));
        return menu;
    }

    /**
     * Creates the font style menu with entries (Plain, Italic, Bold).
     */
    protected Menu createFontStyleMenu() {
        CommandMenu menu = createCommandMenu("Font Style");
        menu.add(new ChangeAttributeCommand("Plain", "FontStyle", new Integer(Font.PLAIN), this));
        menu.add(new ChangeAttributeCommand("Italic","FontStyle", new Integer(Font.ITALIC),this));
        menu.add(new ChangeAttributeCommand("Bold",  "FontStyle", new Integer(Font.BOLD),  this));
        menu.add(new ChangeAttributeCommand("Underlined", "FontStyle", new Integer(TextFigure.Font_UNDERLINED),  this));
        return menu;
    }

    /**
     * Creates the font size menu.
     */
    protected Menu createFontSizeMenu() {
        CommandMenu menu = createCommandMenu("Font Size");
        int sizes[] = { 9, 10, 12, 14, 18, 24, 36, 48, 72 };
        for (int i = 0; i < sizes.length; i++) {
            menu.add(
                new ChangeAttributeCommand(
                    Integer.toString(sizes[i]),
                    "FontSize",
                    new Integer(sizes[i]),  this)
                );
        }
        return menu;
    }

    /**
     * Creates the tool palette.
     */
    protected Panel createToolPalette() {
        Panel palette = new Panel();
        palette.setBackground(Color.lightGray);
        palette.setLayout(new PaletteLayout(2,new Point(2,2),false));
        return palette;
    }

    /**
     * Creates the tools. By default all standard figure creation tools are added.
     * Override this method to remove the standard tools.
     * Call the inherited method to include the standard tools.
     * @param palette the palette where the tools are added.
     */
    protected void createTools(Panel palette) {
        Tool tool = createSelectionTool();

        fDefaultToolButton = createToolButton(IMAGES+"SEL", "Selection Tool", tool);
        palette.add(fDefaultToolButton);

        createAdditionalTools(palette);

        tool = new CreationTool(this, new RectangleFigure());
        palette.add(createToolButton(IMAGES+"RECT", "Rectangle Tool", tool));

        tool = new CreationTool(this, new RoundRectangleFigure());
        palette.add(createToolButton(IMAGES+"RRECT", "Round Rectangle Tool", tool));

        tool = new CreationTool(this, new EllipseFigure());
        palette.add(createToolButton(IMAGES+"ELLIPSE", "Ellipse Tool", tool));

        tool = new CreationTool(this, new DiamondFigure());
        palette.add(createToolButton(IMAGES+"DIAMOND", "Diamond Tool", tool));

        tool = new CreationTool(this, new TriangleFigure());
        palette.add(createToolButton(IMAGES+"TRIANGLE", "Triangle Tool", tool));

        tool = new CreationTool(this, new LineFigure());
        palette.add(createToolButton(IMAGES+"LINE", "Line Tool", tool));

        tool = new ConnectionTool(this, new LineConnection());
        palette.add(createToolButton(IMAGES+"CONN", "Connection Tool", tool));

        tool = new ConnectionTool(this, new ElbowConnection());
        palette.add(createToolButton(IMAGES+"OCONN", "Elbow Connection Tool", tool));

        tool = new ScribbleTool(this);
        palette.add(createToolButton(IMAGES+"SCRIBBL", "Scribble Tool", tool));

        tool = new PolygonTool(this);
        palette.add(createToolButton(IMAGES+"POLYGON", "Polygon Tool", tool));

        tool = new ImageFigureCreationTool(this);
        palette.add(createToolButton(IMAGES+"IMAGE", "Image Tool", tool));

        tool = new TextTool(this, new TextFigure(false));
        fTextTB = createToolButton(IMAGES+"TEXT", "Text Tool", tool);
        palette.add(fTextTB);

        tool = new ConnectedTextTool(this, new TextFigure());
        fConnTextTB = createToolButton(IMAGES+"ATEXT","Connected Text Tool",tool);
        palette.add(fConnTextTB);

    }


    /**
     * Override this method to add additional tools.
     */
    protected void createAdditionalTools(Panel palette) { }


    /**
     * Creates the selection tool used in this editor. Override to use
     * a custom selection tool.
     */
    protected Tool createSelectionTool() {
        return new SelectionTool(this);
    }

    /**
     * Creates a tool button with the given image, tool, and text
     */
    public ToolButton createToolButton(String iconName, String toolName, Tool tool) {
        return new ToolButton(this, iconName, toolName, tool);
    }

    /**
     * Creates the drawing view used in this application.
     * You need to override this method to use a DrawingView
     * subclass in your application. By default a standard
     * DrawingView is returned.
     */
    protected StandardDrawingView createDrawingView(int width, int height) {
        return new StandardDrawingView(this, width, height);
    }

    void unsetCurrentDrawing() { 
      if (fViewFrame!=null) {
	int pos = fDrawings.indexOf(drawing());
	if (pos >= 0) {
	  ((CheckboxMenuItem)fDrawingsMenu.getItem(pos)).setState(false);
	}
	fViewFrame=null;
      }
    }

    void setCurrentDrawing(DrawingViewFrame viewFrame) {
      if (viewFrame==null) {
	unsetCurrentDrawing();
      } else {
	Drawing frameDrawing=viewFrame.view().drawing();
//System.out.println("Trying to activate drawing "+frameDrawing.getName());
	int pos=fDrawings.indexOf(frameDrawing);
	
	// Is the drawing still active or do we process a delayed event?
	if (pos>=0 && fViewFrames.removeElement(viewFrame)) {
          fViewFrames.addElement(viewFrame);
	  unsetCurrentDrawing();
          // new: move viewFrame to last position in vector
          // end new
	  fViewFrame = viewFrame;
	  ((CheckboxMenuItem)fDrawingsMenu.getItem(pos)).setState(true);
//System.out.println("Activated drawing "+frameDrawing.getName());
//displayZOrder();
	}
      }
    }

/* *****
private void displayZOrder() {
   for (int i=0; i<fViewFrames.size(); ++i) {
      System.out.println(i+". "+((DrawingViewFrame)fViewFrames.elementAt(i)).view().drawing().getName());
   }
}
***** */

    protected DrawingViewFrame getViewFrame(Drawing drawing) {
      Enumeration viewFrames = fViewFrames.elements();
      while (viewFrames.hasMoreElements()) {
	  DrawingViewFrame viewFrame =
            (DrawingViewFrame)viewFrames.nextElement();
	  if (viewFrame.view().drawing() == drawing)
                return viewFrame;
       }
       return null;
    }

    public DrawingViewFrame showDrawingViewFrame(Drawing drawing) {
        DrawingViewFrame viewFrame = getViewFrame(drawing);
        if (viewFrame != null) {
            setCurrentDrawing(viewFrame);
            viewFrame.show();
            viewFrame.requestFocus(); // was missing!!!
        } else {
          // Nothing was found. Create a new frame.
          viewFrame = newDrawingViewFrame(drawing);
        }
        return viewFrame;
    }

    public boolean closeDrawing(Drawing drawing) {
        DrawingViewFrame viewFrame = getViewFrame(drawing);
        if (viewFrame != null) {
           return closeViewFrame(viewFrame);
        }
        return false;
    }

    /**
     * Override to define the dimensions of the drawing view.
     */
    protected Dimension getDrawingViewSize() {
        return new Dimension(400, 600);
    }

    /**
     * Creates the drawing used in this application.
     * You need to override this method to use a Drawing
     * subclass in your application. By default a standard
     * Drawing is returned.
     */
    protected Drawing createDrawing() {
        return new StandardDrawing();
    }

    /**
     * Creates the status line.
     */
    protected TextField createStatusLine() {
        TextField field = new TextField("No Tool", 40);
        field.setEditable(false);
        return field;
    }

    /**
     * Handles a user selection in the palette.
     * @see PaletteListener
     */
    public void paletteUserSelected(PaletteButton button) {
        ToolButton toolButton = (ToolButton) button;
        setSelected(toolButton);
        setTool(toolButton.tool(), toolButton.name());
    }

    /**
     * Handles when the mouse enters or leaves a palette button.
     * @see PaletteListener
     */
    public void paletteUserOver(PaletteButton button, boolean inside) {
        ToolButton toolButton = (ToolButton) button;
        if (inside)
            showStatus(toolButton.name());
        else if (fSelectedToolButton != null)
            showStatus(fSelectedToolButton.name());
    }

    /**
     * Gets the current drawing.
     * @see DrawingEditor
     */
    public Drawing drawing() {
        return view().drawing();
    }

    /**
     * Gets the current tool.
     * @see DrawingEditor
     */
    public Tool tool() {
        return fTool;
    }

    /**
     * Gets the current drawing view.
     * @see DrawingEditor
     */
    public DrawingView view() {
      DrawingViewFrame viewFrame=fViewFrame;
      if (viewFrame==null) {
	return NullDrawingView.INSTANCE;
      } else {
	return viewFrame.view();
      }
    }

    /**
     * Gets the drawing view that was active before the current drawing view.
     * @see DrawingEditor
     */
    public DrawingView previousView() {
      int numViews=fViewFrames.size();
      if (numViews >= 2) {
	  return ((DrawingViewFrame)fViewFrames.elementAt(numViews-2)).view();
      }
      return null;
    }

    /**
     * Toggle the sticky mode.
     */
    public void toggleToolButtonSticky() {
      fToolButtonSticky=!fToolButtonSticky;
    }

    /**
     * Sets the default tool of the editor.
     * @see DrawingEditor
     */
    public void toolDone() {
      // not while initializing!
      if (windowCount > 0) {
	if (fToolButtonSticky && tool()!=null) {
	  tool().deactivate();
	  tool().activate();
	} else if (fDefaultToolButton!=null) {
	  setTool(fDefaultToolButton.tool(), fDefaultToolButton.name());
	  setSelected(fDefaultToolButton);
        }
      }
    }

    /**
     * Handles a change of the current selection. Updates all
     * menu items that are selection sensitive.
     * @see DrawingEditor
     */
    public void selectionChanged(DrawingView view) {
        MenuBar mb = menuFrame.getMenuBar();
        CommandMenu editMenu = (CommandMenu)mb.getMenu(EDIT_MENU);
        editMenu.checkEnabled();
        CommandMenu alignmentMenu = (CommandMenu)mb.getMenu(ALIGNMENT_MENU);
        alignmentMenu.checkEnabled();
    }

    /**
     * Shows a status message.
     * @see DrawingEditor
     */
    public void showStatus(String string) {
        fStatusLine.setText(string);
    }

    private void setTool(Tool t, String name) {
        if (fTool != null)
            fTool.deactivate();
        fTool = t;
        if (fTool != null) {
            fStatusLine.setText(name);
            fTool.activate();
        }
    }

    private void setSelected(ToolButton button) {
        if (fSelectedToolButton != null)
            fSelectedToolButton.reset();
        fSelectedToolButton = button;
        if (fSelectedToolButton != null)
            fSelectedToolButton.select();
    }


    public ToolButton toolButtonForTextFigure(TextFigure figure) {
       if (figure.parent() == null) {
          return fTextTB;
       } else {
          return fConnTextTB;
       }
    }

    public void doTextEdit(TextFigure textFigure, int line, int column) {
       ToolButton tb=toolButtonForTextFigure(textFigure);
       paletteUserSelected(tb);
       TextTool tt = (TextTool)tb.tool();
       tt.beginEdit(textFigure);
       if (line > 0 && column > 0) {
	    tt.setCaretPosition(line,column);
       }
    }

    public void doTextEdit(TextFigure textFigure) {
       doTextEdit(textFigure, 0,0);
    }

    /**
     * Exits the application. You should never override this method
     */
    public void exit() {
        // Close all DrawingViewFrames:
        while (true) {
            Enumeration viewFrames = fViewFrames.elements();
            if (!viewFrames.hasMoreElements()) {
               break;
            } else {
              DrawingViewFrame viewFrame =
                 (DrawingViewFrame)viewFrames.nextElement();
              if (!closeViewFrame(viewFrame))
                  return;
            }
        }
        destroy();
        toolFrame.setVisible(false);
        toolFrame.dispose();
        menuFrame.setVisible(false);      // hide the Frame
        menuFrame.dispose();   // tell windowing system to free resources
        if (--windowCount <= 0)
           System.exit(0);
    }

    /**
     * Handles additional clean up operations. Override to destroy
     * or release drawing editor resources.
     */
    protected void destroy() {
    }



    /**
     * Creates a new drawing view with a new empty drawing.
     */
    public void promptNew() {
       Drawing drawing = createDrawing();
       newDrawingViewFrame(drawing);
    }

    /**
     * Shows a file dialog and opens a drawing.
     */
    public void promptOpen() {
      String path = getLoadPath("Open File...", 
				null,
				new String[]{drawExtension,"ser"});
      if (path != null) {
	Drawing drawing;
	loadAndOpenDrawing(path);
      }
    }

   public boolean saveDrawing(Drawing drawing) {
      String filename=drawing.getFilename();
      if (filename==null) {
         if (!promptSaveAs(drawing))
            return false;
         filename=drawing.getFilename();
      }
      saveDrawing(drawing,filename);
      return true;
   }

   public void saveDrawing(Drawing drawing, String filename) {
         String text=" drawing "+drawing.getName()+" as "+filename;
         showAndPrint("Saving"+text+"...");
         if (drawExtension.equals(StringUtil.getExtension(filename))) {
            saveAsStorableOutput(drawing,filename);
         } else {
            saveAsObjectOutput(drawing,filename);
         }
         showStatus("Saved"+text+".");
    }

    /**
     * Shows a file dialog for saving a drawing.
     * Returns null or a complete file name.
     * If a file name is determined, the drawing's name is set accordingly.
     */
    boolean promptSaveAs(Drawing drawing) {
        String filename=drawing.getFilename();
        if (filename==null) {
           filename=fLastPath
	       +File.separator+drawing.getName()
	       +"."+drawExtension;
	}
        toolDone();
        filename = getSavePath("Save File...",
			       filename,
			       new String[]{drawExtension});
			       // was: {drawExtension,"ser"}
        if (filename == null)
           return false;
        updateName(drawing,filename);
        return true;
    }

    /**
     * Prints the drawing.
     */
    public void print() {
        fTool.deactivate();
        PrintJob printJob = menuFrame.getToolkit().getPrintJob(fViewFrame, "Print Drawing", null);

        if (printJob != null) {
            Graphics pg = printJob.getGraphics();

            if (pg != null) {
	        // OK: This was 
	        // fViewFrame.drawingView.printAll(pg);
	        // which relied on a standard drawing view.
	        // There is no reason why a call to the standard
	        // paint method should not succeed.
	        view().paint(pg);
                pg.dispose(); // flush page
            }
            printJob.end();
        }
        fTool.activate();
    }

    /**
     * Opens a save file dialog according to the parameters
     * and returns the selected/entered file name.
     * All parameters are handled as described in<code>
     * getLoadOrSavePath()</code>.
     * @see #getLoadOrSavePath
     **/
    public String getSavePath(String title, 
				 String defaultName,
				 String[] defaultExtensions) {
      return getLoadOrSavePath(title,defaultName,defaultExtensions,FileDialog.SAVE);
    }

    /**
     * Opens a load file dialog according to the parameters
     * and returns the selected/entered file name.
     * All parameters are handled as described in<code>
     * getLoadOrSavePath()</code>.
     * @see #getLoadOrSavePath
     **/
    public String getLoadPath(String title, 
				 String defaultName,
				 String[] defaultExtensions) {
      return getLoadOrSavePath(title,defaultName,defaultExtensions,FileDialog.LOAD);
    }

    /**
     * Opens a file dialog according to the parameters
     * and returns the selected/entered file name.
     * @param title  Title to display in the dialog.
     * @param defaultName  File name to propose to the user.
     *               The file name is checked and completed
     *               to end with one of the default Extensions.
     *               If the file name is prefixed with a
     *               directory, the dialog initially switches
     *               its actual directory to this prefix.
     *               In the other case, the actual directory is
     *               set to the last directory used for files
     *               with the file's extension.
     * @param defaultExtensions  Array of possible extensions.
     *               The file filter is set up to show only
     *               files with one of these extensions.
     *               See <code>checkAndAddExtension()</code>
     *               about the way the default extensions
     *               affect proposed and returned file names.
     *               If no default extensions are specified
     *               (<code>null</code> or empty array), all
     *               files are shown and no extension is required.
     * @param loadOrSave  One of the <code>LOAD</code> or
     *               <code>SAVE</code> constants from
     *               <code>java.awt.FileDialog</code>.
     *               Use <code>getLoadPath()</code> or
     *               <code>getSavePath()</code> to avoid
     *               the need for these constants.
     * @return       Full path name or <code>null</code>, if
     *               the dialog was cancelled.
     *               The path is checked and completed to
     *               end with one of the default extensions
     *               before it is returned - the user has no
     *               possibility to avoid this check.
     * @see #getLoadPath
     * @see #getSavePath
     * @see #checkAndAddExtension
     * @see java.awt.FileDialog
     **/
    protected String getLoadOrSavePath(String title, 
				       String defaultName,
				       String[] defaultExtensions,
				       int loadOrSave) {
        String path = null;
	String defaultDirectory = null;
        FileDialog dialog = new FileDialog(menuFrame, title, loadOrSave);
	if (defaultName != null) {
	  //System.out.println("Given file name:    "+defaultName); 
	  defaultName = checkAndAddExtension(defaultName, defaultExtensions);
	  //System.out.println("After ext. check:   "+defaultName); 
	  defaultDirectory = StringUtil.getPath(defaultName);
	  //System.out.println("Given dir. name:    "+defaultDirectory); 
	  if ("".equals(defaultDirectory)) {
	    defaultDirectory = getDirForExt(StringUtil.getExtension(defaultName));
	    //System.out.println("Looked up dir.:     "+defaultDirectory); 
	  } else {
	    defaultName = StringUtil.getExtendedFilename(defaultName);
	    //System.out.println("Short file name:    "+defaultName); 
	  }
	} else if (defaultExtensions != null) {
	  if (defaultExtensions.length > 0) {
	    //System.out.println("No default name."); 
	    //System.out.println("First default ext.: "+defaultExtensions[0]); 
	    defaultDirectory = getDirForExt(defaultExtensions[0]);
	    //System.out.println("Looked up dir.:     "+defaultDirectory); 
	  }
	}

	if (defaultDirectory != null)
	  dialog.setDirectory(defaultDirectory);
	if (defaultName != null)
	  dialog.setFile(defaultName);
	// Commented out. Might screw Macs.
	//if (defaultExtensions!=null) {
	//  dialog.setFilenameFilter(
	//         new FileByExtensionFilter(defaultExtensions));
	//}

        dialog.show();

        String filename = dialog.getFile();
        if (filename != null) {
            filename = stripTrailingAsterisks(filename);
	    //System.out.println("Entered file name:  "+filename); 
            String dirname = dialog.getDirectory();
	    //System.out.println("Selected directory: "+filename); 
	    if (!dirname.endsWith(File.separator)) {
		dirname += File.separator;
	    }
            path = dirname + filename;
	    path = checkAndAddExtension(path, defaultExtensions);
	    path = StringUtil.makeCanonical(path);
	    //System.out.println("Complete file name: "+path); 
	    updateDirByExt(path);
        }
        dialog.dispose();
        return path;
    }

    /**
     * Checks if the given path ends with one of the given
     * extensions and adds one, if this was not the case.
     * @param path   Filename to check (and modify).
     * @param defaultExtensions  Array of possible extensions.
     *               If the given path has none of these extensions
     *               appended to its end, the first default
     *               extension is automatically appended.
     *               The given extension list is enforced, e.g.
     *               the result is guaranteed to end with one
     *               of these extensions.
     *               If no default extensions are specified
     *               (<code>null</code> or empty array), 
     *               no extension is required or appended.
     * @return       Completed filename (with one of the
     *               given extensions).
     **/
    protected String checkAndAddExtension(String path, 
					  String[] defaultExtensions) {
      if (defaultExtensions != null) {
	boolean hasExtension = false;
	String ext=StringUtil.getExtension(path);
	for (int i = 0; i < defaultExtensions.length; i++)
	  if (defaultExtensions[i].equals(ext))
	    hasExtension = true;
	String result = path;
	if ((defaultExtensions.length > 0) && !hasExtension)
	  result += "."+defaultExtensions[0];
	return result;
      } else {
	return path;
      }
    }

    protected String stripTrailingAsterisks(String filename) {
        // workaround for bug on NT
        if (filename.endsWith("*.*"))
            return filename.substring(0, filename.length() - 4);
        else
            return filename;
    }

    /**
     * Updates the directory-by-extension lookup table by
     * extracting path and extension from the given filename.
     * Then behaves like updateDirByExt(String ext, String path).
     **/
    protected void updateDirByExt(String fullFilename) {
      updateDirByExt(StringUtil.getExtension(fullFilename),
		     StringUtil.getPath(fullFilename));
    }

    /**
     * Updates the directory-by-extension lookup table.
     * Sets the default directory for the given extension
     * to the given path.
     * Also sets the fallback directory to the given path.
     **/
    protected void updateDirByExt(String ext, String path) {
      dirByExt.put(ext, path);
      fallbackDir = path;
    }

    /**
     * Queries the directory-by-extension lookup table
     * for the default directory for the given extension.
     * If the table does not contain an entry for the
     * given extension, returns the fallback directory.
     * May return <code>null</code> if the fallback
     * directory was not set, either.
     **/
    protected String getDirForExt(String ext) {
      String path = (String)dirByExt.get(ext);
      if (path == null) 
	path = fallbackDir;
      return path;
    }
  
	  // Changes version 0->1:
	  //  - PolyLineFigure is now an AttributeFigure.
	  // Changes version 4->5:
	  //  - Saving parameters of arrow tips.
          // Changes version 5->6:
          //  - FSFigures are now CPNTextFigures.
    private static final int FILEVERSION=6;

    private void saveAsStorableOutput(Drawing drawing, String filename) {
        // TBD: should write a MIME header
        try {
            StorableOutput output = new StorableOutput(filename);
	    // Write the current version. This is a
	    // substitute for a MIME header.
	    output.writeInt(FILEVERSION);
	    // Write the drawing.
            output.writeStorable(drawing);
	    // New in 1.2 beta 5 (version 4): Write Window Position:
	    DrawingViewFrame viewFrame=getViewFrame(drawing);
	    Point loc=viewFrame.getLocation();
	    Dimension siz=viewFrame.getSize();
	    output.writeInt(loc.x);
	    output.writeInt(loc.y);
	    output.writeInt(siz.width);
	    output.writeInt(siz.height);
            output.close();
        } catch (IOException e) {
            showStatus(e.toString());
        }
    }

    private void saveAsObjectOutput(Drawing drawing, String file) {
        // TBD: should write a MIME header
        try {
            FileOutputStream stream = new FileOutputStream(file);
            ObjectOutput output = new ObjectOutputStream(stream);
            output.writeObject(drawing);
            output.close();
        } catch (IOException e) {
            showStatus(e.toString());
        }
    }

    private Point newWindowLoc;
    private Dimension newWindowDim;

    public synchronized void loadAndOpenDrawing(String file) {
	newWindowLoc=null;
	newWindowDim=null;
	Drawing drawing  = loadDrawing(file);
	if (drawing != null) {
	    DrawingViewFrame viewFrame=newDrawingViewFrame(drawing);
	    if (newWindowLoc!=null && newWindowDim != null) {
		// restore old window position
		viewFrame.setLocation(newWindowLoc);
		viewFrame.setSize(newWindowDim);
		viewFrame.validate();
	    }
	}
    }

    private void showAndPrint(String msg) {
	showStatus(msg);
	System.out.println(msg);
    }

    private Drawing loadDrawing(String pathname) {
	File file;
	String msg;
	pathname = StringUtil.makeCanonical(pathname);
        showAndPrint("Loading drawing from "+pathname+"...");
        Drawing drawing = null;
//        toolDone();
        String type = guessType(pathname);
        if (type.equals("storable")) {
	    drawing = readFromStorableInput(pathname);
	    showStatus("Loaded drawing from "+pathname+".");
	} else if (type.equals("serialized")) {
            drawing = readFromObjectInput(pathname);
	    showStatus("Loaded serialized drawing from "+pathname+".");
        } else {
	    showAndPrint("Unknown file type of file \""+pathname+"\".");
        }
        if (drawing != null) {
	   drawing.setName(StringUtil.getFilename(pathname));
           drawing.setFilename(pathname);
           fLastPath = StringUtil.getPath(pathname);
           if (drawing instanceof Figure) {
              ((Figure)drawing).invalidate();
           }
	   // As it is possible to load drawings without having
	   // previously used the file dialog, update the directory
	   // lookup table here, too.
	   updateDirByExt(pathname);
        }
        return drawing;
    }

    /**
     * Override to establish a new policy to create
     * input handlers.
     */

    protected StorableInput makeStorableInput(String file, boolean useUFT)
	throws FileNotFoundException {
      return new StorableInput(file,useUFT);
    }

    private Drawing readFromStorableInput(String filename) {
        Drawing drawing = null;
        try {
            StorableInput input = makeStorableInput(filename,true);
            try {
              drawing = readStorableDrawing(input);
            } catch (sun.io.MalformedInputException e) {
               input.close();
               input = makeStorableInput(filename,false);
               drawing = readStorableDrawing(input);
            }
	    try {
		// is there a window position to restore?
		newWindowLoc=new Point(input.readInt(),input.readInt());
		newWindowDim=new Dimension(input.readInt(),input.readInt());
		//System.out.println("Window Coords: "
		//                   +newWindowLoc+" / "+newWindowDim);
	    } catch (Exception e) {}
        } catch (IOException e) {
            showStatus("Error: " + e);
            System.out.println("Error: " + e);
        }
        return drawing;
    }

    private Drawing readStorableDrawing(StorableInput input)
      throws IOException
    {
      if (input.canReadInt()) {
	input.setVersion(input.readInt());
      } else {
	input.setVersion(0);
      }
      
      return (Drawing)input.readStorable();
    }

    private Drawing readFromObjectInput(String file) {
        Drawing drawing = null;
        try {
            FileInputStream stream = new FileInputStream(file);
            ObjectInput input = new ObjectInputStream(stream);
            drawing = (Drawing)input.readObject();
        } catch (IOException e) {
            showStatus("Error: " + e);
            System.out.println("Error: " + e);
        } catch (ClassNotFoundException e) {
            showStatus("Class not found: " + e);
            System.out.println("Error: " + e);
        }
        return drawing;
    }

    private String guessType(String file) {
	String ext=StringUtil.getExtension(file);
        if (drawExtension.equals(ext))
            return "storable";
        if ("ser".equals(ext))
            return "serialized";
        return "unknown";
    }

    public void windowOpened(WindowEvent e) {
       //DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
       //fViewFrames.addElement(viewFrame);
       //System.out.println("New Drawing View Window: "+viewFrame.getTitle());
    }

    public void windowActivated(WindowEvent e) {
      DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
      if (viewFrame != null && viewFrame != fViewFrame) {
	//System.out.println("Current Drawing View Window: "+viewFrame.getTitle());
      }
      setCurrentDrawing(viewFrame);
    }

    public void windowDeactivated(WindowEvent e) {    }

    public void windowClosing(WindowEvent e) {
      DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
      // !!!!!!!!!!! This was:
      // removeDrawing(viewFrame.drawingview.drawing());
      closeViewFrame(viewFrame);
    }

    public void windowClosed(WindowEvent e) {
    }

    public void windowIconified(WindowEvent e) {
    }

    public void windowDeiconified(WindowEvent e) {
    }

    //-- main -----------------------------------------------------------

    public static void main(String[] args) {
	new DrawApplication("JHotDraw",args);
    }

}

class DrawingMenuListener implements ItemListener {
   private DrawApplication fEditor;
   private Drawing fDrawing;

   DrawingMenuListener(DrawApplication editor, Drawing drawing) {
      fEditor = editor;
      fDrawing = drawing;
   }

   public void itemStateChanged(ItemEvent event) {
      fEditor.showDrawingViewFrame(fDrawing);
   }
}

class DrawingLoadServer extends Thread {

        private DrawApplication editor;
        private ServerSocket s=null;

        public DrawingLoadServer(DrawApplication editor, int port) throws IOException {
            this.editor=editor;
            System.out.println("Drawing Load Server setting up server socket at port "+port+"...");
            s=new ServerSocket(port);
        }

        public void run() {
            System.out.println("Drawing Load Server waiting for parameters...");
            while (true) {
              Socket client=null;
              try {
                 client=s.accept();
                 System.out.println("Parameter server accepted client.");
                 BufferedReader in=new BufferedReader(new InputStreamReader(client.getInputStream()));
                 try {
                    String drawingFileName;
                    do {
                         drawingFileName=in.readLine();
                         if (drawingFileName!=null) {
                            System.out.println("Received Parameter "+drawingFileName);
                            editor.loadAndOpenDrawing(drawingFileName);
                         }
                    } while (drawingFileName != null);
                 } catch (IOException e) { }
                 System.out.println("Connection closed.");
                 in.close();
                 client.close();
              } catch (Exception e) {
                 System.out.println("Drawing Load Server threw exception: "+e);
              }
           }
        }
}
