package de.renew.database;

import java.util.Vector;
import de.renew.simulator.PlaceInstance;

/**
 * An object of this type may be used to extract the current information
 * about a system of nets out of a persistent store.
 *
 * While this information is read, it can be assumed that no changes
 * are performed to the database. On the other hand, it may be more efficient
 * to read the entire database into this object and use cached
 * data during the query. 

 * This accessor interface suggests the following structure of the
 * database, but otherstructures are possible, too.
 *
 * table NETS: rows NAME, NETID
 * table MARKING: rows NETID, PLACE, TYPE, TOKEN, COUNT
 *
 * It is suggested that the type of the token is saved, too,
 * in order to be able to correctly interpret the
 * stored token object. It may be required to store
 * different token types in different tables, if the structure of the
 * tokens is very irregular.
 *
 * References to net instances should be saved as the net instance's ID
 * with an appropriate type. Upon restoring the net instances,
 * a map will be provided that allows to infer net instances from IDs.
 */
public interface RestoreSource {
  /**
   * Get all net identifiers as strings.
   *
   * @return an array of all net IDs
   */
  public String[] getAllNetIDs();

  /**
   * Translates a net identifier into a net name.
   * The net name is typically the name given to the net file.
   *
   * @param netID net identifier
   * @return symbolic name associated with the identifier
   */
  public String getNetName(String netID);

  /**
   * Get all the tokens which exist at a place in a net.
   *
   * If the place contains references to net instances,
   * these have been typically saved by their net ID only.
   * A special lookup table simplifies the lookup of net instances.
   *
   * @see de.renew.simulator.NetInstance
   *
   * If the implementing class needs to access the
   * IDs that characterize this place instance, it is
   * suggested to use a code fragement like the following:
   * <code><pre>
   * String netID=placeInstance.getNetInstance().getID();
   * int placeID=placeInstance.getPlace().getID();
   * </pre></code>
   *
   * @param placeInstance the place instance where the token is
   *   added
   * @param map a hashtable that allows the database to look up
   *   references to net instances, if given the nets ID
   * @param ids a vector, to which the implementing class must
   *   append all IDs of token in this place
   * @param tokens a vector, to which the implementing class must
   *   append all values of token in this place in the same order
   *   that was used for the ids
   */
  public void fillinAllTokens(PlaceInstance placeInstance,
			      NetInstanceMap map,
			      Vector ids, Vector tokens);
}
