package de.renew.database;

import java.util.Vector;
import de.renew.simulator.PlaceInstance;
import de.renew.simulator.NetInstance;

/**
 * Accept tasks for a transaction and make sure to forward
 * them to the transaction strategy in due time. Additionally,
 * take a note of all token deposits and perform them after the
 * transaction was committed.
 *
 * This enables us to delay the actual token output and hence
 * the firing of further transitions until the firing of
 * the first transition is firmly recorded in the database.
 *
 * For any token that is added or removed, transactions
 * are also responsible for unreserving that token from the
 * place instance once the final fate of the transaction
 * (commit or rollback) has been decided and no more actions are
 * to be performed.
 */
public class Transaction {
  private boolean autoCommit;

  private Vector createActions;
  private Vector deleteActions;
  private Vector addActions;
  private Vector removeActions;

  private Vector placeInstances;
  private Vector tokens;
  private Vector times;

  /**
   * Create a transaction that records all calls and commits them
   * jointly in the end or a transaction that immediately commits each
   * modification.
   *
   * @param autoCommit true, if all actions should be immediately committed
   */
  public Transaction(boolean autoCommit) {
    this.autoCommit=autoCommit;

    createActions=new Vector();
    deleteActions=new Vector();
    addActions=new Vector();
    removeActions=new Vector();

    placeInstances=new Vector();
    tokens=new Vector();
    times=new Vector();
  }

  public synchronized void createNet(NetInstance instance) {
    createActions.addElement(new NetAction(instance));

    if (autoCommit) {
      commit();
    }
  }

  public synchronized void deleteNet(NetInstance instance) {
    deleteActions.addElement(new NetAction(instance));

    if (autoCommit) {
      commit();
    }
  }

  public synchronized void addToken(PlaceInstance placeInstance, Object token,
				    double time, boolean automaticInsertion)
  {
    // If the insertion of the token happens under the transaction's control,
    // make sure that its ID does not change.
    if (automaticInsertion) {
      placeInstance.reserve(token);
    }

    addActions.addElement(new TokenAction(placeInstance,token));

    if (automaticInsertion) {
      placeInstances.addElement(placeInstance);
      tokens.addElement(token);
      times.addElement(new Double(time));
    }

    if (autoCommit) {
      commit();
    }
  }

  public synchronized void removeToken(PlaceInstance placeInstance,
				       Object token)
  {
    placeInstance.reserve(token);

    removeActions.addElement(new TokenAction(placeInstance,token));

    if (autoCommit) {
      commit();
    }
  }

  private void outputTokens() {
    int n=placeInstances.size();
    for (int i=0;i<n;i++) {
      PlaceInstance pi=(PlaceInstance)placeInstances.elementAt(i);
      Object token=tokens.elementAt(i);
      double time=((Double)times.elementAt(i)).doubleValue();
      pi.lock.lock();
      pi.internallyInsertToken(token,time);
      pi.lock.unlock();
    }
  }

  private void unreserveOutputTokens() {
    int n=placeInstances.size();
    for (int i=0;i<n;i++) {
      PlaceInstance pi=(PlaceInstance)placeInstances.elementAt(i);
      Object token=tokens.elementAt(i);
      pi.lock.lock();
      pi.unreserve(token);
      pi.lock.unlock();
    }
  }

  private void unreserveInputTokens() {
    int n=removeActions.size();
    for (int i=0;i<n;i++) {
      TokenAction tokenAction=(TokenAction)removeActions.elementAt(i);
      PlaceInstance pi=tokenAction.getPlaceInstance();
      pi.lock.lock();
      pi.unreserve(tokenAction.getToken());
      pi.lock.unlock();
    }
  }

  private void clear() {
    placeInstances.removeAllElements();
    tokens.removeAllElements();
    times.removeAllElements();
  }

  public synchronized void commit() {
    NetAction[] ca=new NetAction[createActions.size()];
    createActions.copyInto(ca);
    TokenAction[] aa=new TokenAction[addActions.size()];
    addActions.copyInto(aa);
    TokenAction[] ra=new TokenAction[removeActions.size()];
    removeActions.copyInto(ra);
    NetAction[] da=new NetAction[deleteActions.size()];
    deleteActions.copyInto(da);

    TransactionSource.perform(ca,aa,ra,da);

    outputTokens();
    unreserveOutputTokens();
    unreserveInputTokens();

    clear();
  }

  public synchronized void rollback() {
    unreserveOutputTokens();
    unreserveInputTokens();
    clear();
  }    
}
