package de.renew.database;

/**
 * This interface encapsulates the code that an application
 * has to execute when the results of a firing transition
 * have to written to a database.
 */
public interface TransactionStrategy {
  /**
   * Execute one transaction, e.g., by writing to a database.
   *
   * A typical database might feature these tables:
   * <dl>
   * <dt>NETS</dt><dd>NETID: String, NAME: String</dd>
   * <dt>TOKENS</dt><dd>NETID: String, PLACEID: int, TOKENID: String,
   *   TOKENDATA: byte[]</dd>
   * </dl>
   * The TOKENID column is used to decide equality of the stored
   * objects. the TOKENDATA column might contain the serialized
   * versions of the tokens. Applications may decide to use
   * a places name instead of its ID for database accesses.
   * 
   * It is suggested that the actions are executed in the
   * order of their appearance in the method header.
   * Especially, it is possible that tokens are added in a 
   * newly created net, that newly deposited tokens are
   * removed immediately, and that tokens are removed before
   * a net is finally deleted.
   *
   * This call must succeed. It has been verified by the
   * simulator that all available tokens are in the appropriate
   * places and have been assigned the proper ID.
   * The action objects may be queried for all relevant
   * information about net creations and token moves.
   *
   * @param createActions an array of net creation actions 
   * @param addActions an array of token insertion actions 
   * @param removeActions an array of token removal actions 
   * @param deleteActions an array of net deletion actions 
   */
  public void perform(NetAction[] createActions,
		      TokenAction[] addActions,
		      TokenAction[] removeActions,
		      NetAction[] deleteActions);
}
