package de.renew.formalism.java;

import collections.*;
import de.renew.util.Types;
import de.renew.util.SingletonEnumeration;
import de.renew.util.EmptyEnumeration;
import de.renew.expression.*;
import de.renew.simulator.*;
import de.renew.shadow.*;
import de.renew.unify.*;
import de.renew.function.*;
import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

public class JavaNetCompiler implements ShadowCompiler,
  ImportCompiler
{
  private boolean allowDangerousArcs=false;
  private boolean allowTimeInscriptions=false;
  private boolean wantEarlyTokens=false;

  // My other fields need not be serialized.
  // They are set during a compilation. This holds even 
  // for the shadow net system that I have to compile.
  // This is a bit strange, but it causes no problems in practice.
  static final long serialVersionUID = 6357839474457945975L;

  private transient ShadowLookup lookup;
  private transient Hashtable placeTypes; // Maps shadow places to types.
  private transient ShadowNetSystem netSystem;
  private transient int placeNum=0;
  private transient int transitionNum=0;

  // The no-arg constructor is especially important when an object of
  // this class is created through Class.forName(...).newInstance() .
  public JavaNetCompiler() {
    this(false,false,false);
  }

  public JavaNetCompiler(boolean allowDangerousArcs,
    boolean allowTimeInscriptions, boolean wantEarlyTokens)
  {
    this.allowDangerousArcs=allowDangerousArcs;
    this.allowTimeInscriptions=allowTimeInscriptions;
    this.wantEarlyTokens=wantEarlyTokens;
  }

  protected InscriptionParser makeParser(String inscr) {
    return new JavaNetParser(new StringReader(inscr));
  }

  public static SyntaxException makeSyntaxException(ParseException e) {
    Token t=e.currentToken.next;
    return new SyntaxException(e.getMessage(),null,
      t.beginLine,t.beginColumn);
  }

  public NameInterpreter compileImports(ShadowDeclarationNode declaration)
    throws SyntaxException
  {
    try {
      return parseDeclarationNode(declaration.inscr);
    }
    catch (SyntaxException e) {
      throw e.addObject(declaration);
    }
  }

  // Check the validity of certain inscriptions.
  public String checkDeclarationNode(String inscr, boolean special)
    throws SyntaxException
  {
    parseDeclarationNode(inscr);
    return "declaration";
  }

  public String checkArcInscription(String inscr, boolean special,
				    ShadowNet shadowNet)
    throws SyntaxException
  {
    parseArcInscription(inscr,makeDeclarationNode(shadowNet));
    return "expression";
  }

  public String checkTransitionInscription(String inscr, boolean special,
					   ShadowNet shadowNet)
    throws SyntaxException 
  {
    CollectionEnumeration enum=
      makeInscriptions(inscr,null,false,makeDeclarationNode(shadowNet));

    if (enum.numberOfRemainingElements()!=1) {
      return null;
    }
    
    Object res=enum.nextElement();
    if (res instanceof ActionInscription) {
      return "action";
    } else if (res instanceof GuardInscription) {
      return "guard";
    } else if (res instanceof UplinkInscription) {
      return "uplink";
    } else if (res instanceof DownlinkInscription) {
      return "downlink";
    } else if (res instanceof ExpressionInscription) {
      return "expression";
    } else {
      return null;
    }
  }

  public String checkPlaceInscription(String inscr, boolean special,
				      ShadowNet shadowNet)
    throws SyntaxException 
  {
    ParsedDeclarationNode declNode=makeDeclarationNode(shadowNet);
    CollectionEnumeration enum=parsePlaceInscription(inscr,declNode);

    if (enum.numberOfRemainingElements()!=1) {
      return null;
    }
    
    if (enum.nextElement() instanceof Class) {
      return "type";
    } else {
      return "initialmarking";
    }
  }

  // Do the actual compilation.

  ParsedDeclarationNode compile(ShadowDeclarationNode declaration)
    throws SyntaxException
  {
    try {
      return parseDeclarationNode(declaration.inscr);
    }
    catch (SyntaxException e) {
      throw e.addObject(declaration);
    }
  }

  void compile(ShadowPlace shadowPlace, ParsedDeclarationNode declaration,
    Net net) throws SyntaxException
  {
    // Determine the name.
    String pname=shadowPlace.getName();
    if (pname==null) {
      placeNum++;
      pname="P"+placeNum;
    }

    // Create the new place;
    Place place=new Place(net,pname);
    place.setID(shadowPlace.getID());
    place.setTrace(shadowPlace.getTrace());
    lookup.set(shadowPlace,place);

    // Store the initialisation inscriptions for later use.
    Vector parsedInscriptions=new Vector();
    Vector errorShadows=new Vector();

    // Insert the inscriptions.
    CollectionEnumeration enum=shadowPlace.elements();
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowInscription) {
	String inscr=((ShadowInscription)elem).inscr;

	try {
	  CollectionEnumeration exprEnum=
	    parsePlaceInscription(inscr,declaration);
	  while (exprEnum.hasMoreElements()) {
	    parsedInscriptions.addElement(exprEnum.nextElement());
	    errorShadows.addElement(elem);      
	  }
	} catch (SyntaxException e) {
	  throw e.addObject((ShadowInscription)elem);
	}
      } else {
	throw new SyntaxException("Unsupported place inscription",
	  null).addObject(shadowPlace).addObject(elem);
      }
    }

    // Reember the type of this place.
    Class type=Types.UNTYPED;
    int typeCount=0;

    for (int i=0; i<parsedInscriptions.size(); i++) {
      if (parsedInscriptions.elementAt(i) instanceof Class) {
	type=(Class)parsedInscriptions.elementAt(i);
	typeCount++;
      }
    }

    // Check for multiple types.
    if (typeCount>1) {
      SyntaxException e=new SyntaxException("Place is typed more than once.",
	null);
      for (int i=0; i<parsedInscriptions.size(); i++) {
	if (parsedInscriptions.elementAt(i) instanceof Class) {
	  e.addObject(errorShadows.elementAt(i));
	}
      }
      throw e;
    }

    // Remember the type information for this place.
    if (type!=Types.UNTYPED) {
      placeTypes.put(shadowPlace,type);
    }

    // Now add the initial markings.
    for (int i=0; i<parsedInscriptions.size(); i++) {
      if (parsedInscriptions.elementAt(i) instanceof TypedExpression) {
	TypedExpression expr=(TypedExpression)parsedInscriptions.elementAt(i);
	
	Expression castedExpression=null;
	try {
	  castedExpression=JavaNetHelper.makeCastedOutputExpression(type,
	    expr);
	} catch (SyntaxException e) {
	  throw e.addObject(errorShadows.elementAt(i));
	}
	place.add(new ExpressionTokenSource(castedExpression));
      }
    }
  }

  Class getType(ShadowPlace place) {
    if (placeTypes.containsKey(place)) {
      return (Class)placeTypes.get(place);
    } else {
      return Types.UNTYPED;
    }
  }

  void compile(ShadowTransition shadowTransition,
    ParsedDeclarationNode declaration, Net net) throws SyntaxException 
  {
    // Determine the name.
    String tname=shadowTransition.getName();
    if (tname==null) {
      transitionNum++;
      tname="T"+transitionNum;
    }

    // Create the new transition.
    Transition transition=new Transition(net,tname);
    transition.setID(shadowTransition.getID());
    transition.setTrace(shadowTransition.getTrace());
    lookup.set(shadowTransition,transition);

    // Prepare a list to hold the inscriptions.
    Vector parsedInscriptions=new Vector();
    Vector errorShadows=new Vector();

    // Insert the inscriptions.
    CollectionEnumeration enum=shadowTransition.elements();
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowInscription) {
	ShadowInscription inscription=(ShadowInscription)elem;
	
	CollectionEnumeration inscriptions=null;
	try {
	  inscriptions=makeInscriptions(inscription.inscr,
	    lookup,true,declaration);
	} catch (SyntaxException e) {
	  throw e.addObject(inscription);
	}

	while (inscriptions.hasMoreElements()) {
	  parsedInscriptions.addElement(inscriptions.nextElement());
	  errorShadows.addElement(inscription);
	}
      } else if (elem instanceof ShadowArc) {
	compile((ShadowArc)elem,declaration);
      } else {
	throw new SyntaxException("Unsupported transition inscription",
	  null).addObject(shadowTransition).addObject(elem);
      }
    }

    // Check for multiple uplinks.
    int uplinkCount=0;
    for (int i=0; i<parsedInscriptions.size(); i++) {
      if (parsedInscriptions.elementAt(i) instanceof UplinkInscription) {
	uplinkCount++;
      }
    }

    if (uplinkCount>1) {
       SyntaxException e=new SyntaxException(
         "Transition has more than one uplink.",null);
      for (int i=0; i<parsedInscriptions.size(); i++) {
	if (parsedInscriptions.elementAt(i) instanceof UplinkInscription) {
	  e.addObject(errorShadows.elementAt(i));
	}
      }
      throw e;     
    }

    // Insert inscriptions.
    for (int i=0; i<parsedInscriptions.size(); i++) {
      transition.add(
        (TransitionInscription)parsedInscriptions.elementAt(i));
    }
  }

  protected ShadowInscription getSingleArcInscription(ShadowArc arc) 
    throws SyntaxException
  {
    CollectionEnumeration inscriptions=arc.elements();
    if (inscriptions.numberOfRemainingElements()>1) {
      SyntaxException e=
	new SyntaxException("This arc must not be inscribed "+
	  "multiple times.",null).addObject(arc);
      while (inscriptions.hasMoreElements()) {
	e.addObject((ShadowInscription)inscriptions.nextElement());
      }
      throw e;
    }

    CollectionEnumeration enum=new EmptyEnumeration();
    if (inscriptions.hasMoreElements()) {
      return (ShadowInscription)inscriptions.nextElement();
    } else {
      return null;
    }
  }

  protected TypedExpression parseVariable(String inscr,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    InscriptionParser parser=makeParser(inscr);
    parser.setDeclarationNode(declaration);
    try {
      return parser.VariableInscription();
    }
    catch (ParseException e) {
      throw makeSyntaxException(e);
    }
  }

  protected void compileClearArc(ShadowArc shadowArc,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    Transition transition=lookup.get(shadowArc.transition);
    Place place=lookup.get(shadowArc.place);
    Class placeType=getType(shadowArc.place);

    ShadowInscription shadowInscription=getSingleArcInscription(shadowArc);
    String inscr=null;
    if (shadowInscription!=null) {
      inscr=shadowInscription.inscr;
    }

    if (inscr==null || inscr.equals("")) {
      // For empty clear arcs, prepare an anonymous variable.
      ClearArc arc=new ClearArc(place,
	new VariableExpression(Types.UNTYPED,null),
	Object.class);
      arc.setTrace(shadowArc.getTrace());
      transition.add(arc);	
    } else {
      // We have got a expression to parse.
      TypedExpression expr;
      try {
	expr=parseVariable(inscr,declaration);
      }
      catch (SyntaxException e) {
	throw e.addObject(shadowInscription);
      }
      Class exprType=expr.getType();

      Class elementType;
      if (placeType==Types.UNTYPED) {
	if (exprType!=Types.UNTYPED) {
	  throw new SyntaxException("Cannot clear untyped place using "+
	    "typed variable.",null).addObject(shadowArc);
	}

	// If primitive values happen to occur in the
	// place, they are wrapped in Value objects.
	//
	// This exposes the Value class to users, so maybe we should
	// complain? Or should we disable the transition at runtime,
	// if primitive values occur in the place?
	elementType=Object.class;
      } else {
	if (exprType==Types.UNTYPED) {
	  elementType=placeType;
	} else if (!exprType.isArray()) {
	  throw new SyntaxException("Variable of array type expected.",null)
	    .addObject(shadowInscription);
	} else {
	  elementType=exprType.getComponentType();
	  
	  if (!Types.allowsLosslessWidening(placeType,elementType)) {
	    throw new SyntaxException("Cannot losslessly convert "+
	      JavaHelper.makeTypeErrorString(placeType)+
	      " to "+JavaHelper.makeTypeErrorString(elementType)+".",
	      null).addObject(shadowInscription);
	  }
	}
      }
	
      ClearArc arc=new ClearArc(place,expr.getExpression(),
	elementType);
      arc.setTrace(shadowArc.getTrace());
      transition.add(arc);		    
    }
  }

  // Returns an enumeration of correctly timed expressions.
  // This method does not do a lot. Simply checking for
  // void expressions now. Possible simplification.
  protected CollectionEnumeration makeArcExpressions(String inscription,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    CollectionEnumeration timedExprs=parseArcInscription(inscription,
      declaration);
    UpdatableSeq seq=new LinkedList();

    while (timedExprs.hasMoreElements()) {
      TimedExpression expr=(TimedExpression)timedExprs.nextElement();

      if (expr.getExpression().getType()==Void.TYPE) {
	throw new SyntaxException(
	  "Cannot use void expressions as arc inscriptions.",null);
      }

      seq.insertFirst(expr);
    }

    return seq.elements();
  }
    
  // Returns an enumeration of typed expressions.
  protected CollectionEnumeration makeEmptyArcExpressions(
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    return makeArcExpressions(null,declaration);
  }

  protected void compileSingleArcInscription(CollectionEnumeration exprs,
    Class placeType, ArcFactory factory,
    Place place, Transition transition, boolean trace)
    throws SyntaxException
  {
    while (exprs.hasMoreElements()) {
      TimedExpression timedExpr=(TimedExpression)exprs.nextElement();
      
      if (factory.allowsTime() && allowTimeInscriptions) {
	// Verify that the time expression is numeric.
	
	if (timedExpr.isTimed() && timedExpr.getTime().isTyped()) {
	  if (!Types.allowsWideningConversion(
	    timedExpr.getTime().getType(),Double.TYPE))
	  {
	    throw new SyntaxException("Non-numeric time expression.",null);
	  }
	}
      } else if (timedExpr.isTimed()) {
	if (allowTimeInscriptions) {
	  throw new SyntaxException(
            "This arc type does not allow time inscriptions.",null);
	} else {
	  throw new SyntaxException(
            "Time annotations are not allowed.",null);
	}
      }

      factory.compileArc(place,transition,trace,placeType,timedExpr);
    }
  }

  protected void compileArcInscriptions(ShadowArc shadowArc,
    ArcFactory factory,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    Transition transition=lookup.get(shadowArc.transition);
    Place place=lookup.get(shadowArc.place);

    CollectionEnumeration inscriptions=shadowArc.elements();
    if (inscriptions.hasMoreElements()) {
      do {
	ShadowInscription inscription=
	  (ShadowInscription)inscriptions.nextElement();
	
	// Create an individual arc for each inscription.
	try {
	  compileSingleArcInscription(
	    makeArcExpressions(inscription.inscr,declaration),
            getType(shadowArc.place),factory,
	    place,transition,shadowArc.getTrace());
	} catch (SyntaxException e) {
	  throw e.addObject(inscription);
	}
      } while (inscriptions.hasMoreElements());
    } else {
      // Create a single arc with the default inscription.
      try {
	factory.emptyArcCheck();

	compileSingleArcInscription(
          makeEmptyArcExpressions(declaration),
          getType(shadowArc.place),factory,
          place,transition,shadowArc.getTrace());
      } catch (SyntaxException e) {
	throw e.addObject(shadowArc);
      }
    }
  }

  // Redefine this method to map some or all shadow arc types to
  // different simulator arc types.
  protected void compile(ShadowArc shadowArc,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    switch (shadowArc.shadowArcType) {
    case ShadowArc.test: 
      compileArcInscriptions(shadowArc,new SimpleArcFactory(Arc.test,false),
	declaration);
      return;
    case ShadowArc.both: 
      compileArcInscriptions(shadowArc,new SimpleArcFactory(Arc.bothOT,true),
	declaration);
      return;
    case ShadowArc.ordinary: 
      compileArcInscriptions(shadowArc,
	new SimpleArcFactory(shadowArc.placeToTransition?Arc.in:Arc.out,true),
	declaration);
      return;
    case ShadowArc.inhibitor: 
      if (allowDangerousArcs) {
	compileArcInscriptions(shadowArc,new SimpleArcFactory(Arc.inhibitor,
          false),declaration);
	return;
      }
      break;
    case ShadowArc.doubleOrdinary: 
      if (shadowArc.placeToTransition) {
	compileArcInscriptions(shadowArc,FlexibleInArcFactory.INSTANCE,
	  declaration);
      } else {
	compileArcInscriptions(shadowArc,FlexibleOutArcFactory.INSTANCE,
	  declaration);
      }
      return;
    case ShadowArc.doubleHollow: 
      if (allowDangerousArcs) {
	compileClearArc(shadowArc,declaration);
	return;
      }
      break;
    }
    throw new SyntaxException("Unsupported arc type.",
      null).addObject(shadowArc);
  }

  ParsedDeclarationNode parseDeclarationNode(String inscr)
    throws SyntaxException
  {
    if (inscr!=null) {
      InscriptionParser parser=makeParser(inscr);
      try {
	return parser.DeclarationNode();
      }
      catch (ParseException e) {
	throw makeSyntaxException(e);
      }
    } else {
      return new ParsedDeclarationNode();
    }
  }

  CollectionEnumeration parsePlaceInscription(String inscr,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    if ((inscr!=null) && !inscr.equals("")) {
      InscriptionParser parser=makeParser(inscr);
      parser.setDeclarationNode(declaration);
      try {
	return parser.PlaceInscription();
      }
      catch (ParseException e) {
	// I really do not know what this could be.
	// I'll give out the error message of the
	// second parse, although it might be
	// appropriate to tell the user that I also
	// tried to parse a type.
	throw makeSyntaxException(e);
      }
    }

    return new EmptyEnumeration();
  }

  // Returns an enumeration of typed expressions.
  protected CollectionEnumeration parseArcInscription(String inscr,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    if (inscr!=null && !inscr.equals("")) {
      InscriptionParser parser=makeParser(inscr);
      parser.setDeclarationNode(declaration);
      try {
	return parser.ArcInscription();
      }
      catch (ParseException e) {
	throw makeSyntaxException(e);
      }
    } else {
      return new SingletonEnumeration(new TimedExpression(
        new TypedExpression(Tuple.class,
	  new ConstantExpression(Tuple.class,new Tuple(0))),
        null));
    }
  }

  private CollectionEnumeration makeInscriptions(String str,
    ShadowLookup lookup, boolean create,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    // The input string must be parsed, it denotes either a
    // creation inscription, an uplink, a downlink, ...
    //
    // Syntax:
    //   x:new NetName          creation
    //   x:channel(...)         downlink
    //   :channel(...)          uplink
    //   x=new a.java.Type(...) creation
    //   x=y.method(...)        call
    //   guard ...              guard
    //   action x= ...          action

    InscriptionParser parser=makeParser(str);
    parser.setLookup(lookup);
    parser.setDeclarationNode(declaration);

    try {
      return parser.TransitionInscription(create);
    }
    catch (ParseException e) {
      throw makeSyntaxException(e);
    }
  }  

  ParsedDeclarationNode makeDeclarationNode(ShadowNet shadowNet)
    throws SyntaxException
  {
    // Search for declarations.
    CollectionEnumeration enum=shadowNet.elements();
    ShadowDeclarationNode node=null;
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowDeclarationNode) {
        // Have we already found a declaration node?
        if (node!=null) {
	  SyntaxException e=
	    new SyntaxException("Only one declaration node is allowed.",null);
	  e.addObject(node);
	  e.addObject((ShadowDeclarationNode)elem);
	  while (enum.hasMoreElements()) {
	    elem=enum.nextElement();
	    if (elem instanceof ShadowDeclarationNode) {
	      e.addObject((ShadowDeclarationNode)elem);
	    }
	  }
	  throw e;
	}

	// Remember the declaration node we found.
	node=(ShadowDeclarationNode)elem;
      }
    }

    // Compile the declaration node.
    ParsedDeclarationNode declaration;
    if (node==null) {
      declaration=new ParsedDeclarationNode();
    } else {
      declaration=compile((ShadowDeclarationNode)node);
    }
    return declaration;
  }

  void compile(ShadowNet shadowNet) throws SyntaxException {
    // Compile the individual net elements.
    CollectionEnumeration enum;

    // Find the net that is compiled.
    Net net=lookup.getNet(shadowNet.getName());

    // First the declaration node, if any.
    ParsedDeclarationNode declaration=makeDeclarationNode(shadowNet);

    // Second the places.
    enum=shadowNet.elements();
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowPlace) {
	compile((ShadowPlace)elem,declaration,net);
      }
    }

    // Third the transitions.
    enum=shadowNet.elements();
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowTransition) {
	compile((ShadowTransition)elem,declaration,net);
      }
    }
  }

  void createNets() throws SyntaxException {
    // Create new nets for every shadow net.
    // Check the uniqueness of the names.
    CollectionEnumeration enum=netSystem.elements();
    while (enum.hasMoreElements()) {
      ShadowNet shadowNet=(ShadowNet)enum.nextElement();
      if (lookup.getNet(shadowNet.getName())!=null) {
	throw new SyntaxException("Detected two nets with the same name: "+
	  shadowNet.getName()+".",null);
      }
      Net net=new Net(shadowNet.getName());
      net.setEarlyTokens(wantEarlyTokens);
      lookup.setNet(shadowNet.getName(),net);
    }
  }

  protected void createNetElements() throws SyntaxException {
    CollectionEnumeration enum=netSystem.elements();
    while (enum.hasMoreElements()) {
      compile((ShadowNet)enum.nextElement());
    }
  }

  protected ShadowLookup getLookup() { return lookup; }
  protected ShadowNetSystem getNetSystem() { return netSystem; }

  public synchronized ShadowLookup compile(ShadowNetSystem netSystem)
    throws SyntaxException
  {
        return compile(new ShadowLookup(), netSystem);
   }

  protected synchronized ShadowLookup compile(ShadowLookup lookup, ShadowNetSystem netSystem)
    throws SyntaxException
  {
    this.lookup=lookup;
    placeTypes=new Hashtable();
    this.netSystem=netSystem;
    placeNum=0;
    transitionNum=0;
 
    createNets();
    createNetElements();

    // Enable garbage collection.
    this.lookup=null;
    placeTypes=null;
    this.netSystem=null;

    return lookup;
  }

  protected void channelCheck(ShadowNetSystem netSystem)
    throws SyntaxException
  {
    UpdatableSet spontaneousChannels=new HashedSet();
    Hashtable channelMap=new Hashtable();

    CollectionEnumeration nets=netSystem.elements();
    while (nets.hasMoreElements()) {
      ShadowNet shadowNet=(ShadowNet)nets.nextElement();

      ParsedDeclarationNode declarations=makeDeclarationNode(shadowNet);

      CollectionEnumeration elements=shadowNet.elements();
      while (elements.hasMoreElements()) {
	Object element=elements.nextElement();
	if (element instanceof ShadowTransition) {
	  ShadowTransition transition=(ShadowTransition)element;

	  ChannelInscription uplink=null;
	  UpdatableSet downlinks=new HashedSet();

	  CollectionEnumeration inscriptions=transition.elements();
	  while (inscriptions.hasMoreElements()) {
	    Object inscription=inscriptions.nextElement();
	    if (inscription instanceof ShadowInscription) {
	      String text=((ShadowInscription)inscription).inscr;
	      try {
		InscriptionParser parser=makeParser(text);
		parser.setDeclarationNode(declarations);

		ChannelInscription ch=parser.tryParseChannelInscription();
		if (ch.isUplink) {
		  if (uplink!=null) {
		    throw new SyntaxException(
                      "Transition has more than one uplink.",
		      null).addObject(transition);
		  }
		  uplink=ch;
		} else {
		  downlinks.include(ch);
		}
	      }
	      catch (ParseException e) {
		// This is expected. We parsed something other than
		// a channel. We ignore this inscription.
	      }
	    }
	  }

	  // Reduce all channel inscriptions to downlinks and
	  // spontaneous transitions to invoked transitions.
	  if (uplink==null) {
	    uplink=new ChannelInscription(false,"",0);
	  } else {
	    uplink=new ChannelInscription(false,uplink.name,uplink.arity);
	  }
	  ChannelCheckNode source=(ChannelCheckNode)channelMap.get(uplink);
	  if (source==null) {
	    source=new ChannelCheckNode(uplink.name,uplink.arity);
	    channelMap.put(uplink,source);
	  }

	  CollectionEnumeration channels=downlinks.elements();
	  if (channels.hasMoreElements()) {
	    do {
	      ChannelInscription channel=
		(ChannelInscription)channels.nextElement();
	      ChannelCheckNode target=
                (ChannelCheckNode)channelMap.get(channel);
	      if (target==null) {
		target=new ChannelCheckNode(channel.name,channel.arity);
		channelMap.put(channel,target);
	      }
	    
	      source.addInvokableChannel(target);
	    } while (channels.hasMoreElements());
	  } else {
	    source.setSatisfiable();
	  }
	}
      }
    }

    // Check all channels, even though they might not
    // be accessible.
    Enumeration enum=channelMap.elements();
    while (enum.hasMoreElements()) {
      ChannelCheckNode node=(ChannelCheckNode)enum.nextElement();
      node.check();
    }
  }

  protected void doubleNameCheck(ShadowNetSystem netSystem)
    throws SyntaxException
  {
    CollectionEnumeration nets=netSystem.elements();
    while (nets.hasMoreElements()) {
      ShadowNet shadowNet=(ShadowNet)nets.nextElement();
      
      Hashtable names=new Hashtable();
      CollectionEnumeration elements=shadowNet.elements();
      while (elements.hasMoreElements()) {
	Object element=elements.nextElement();
	if (element instanceof ShadowNode) {
	  ShadowNode node=(ShadowNode)element;
	  String name=node.getName();
	  if (name!=null && !name.equals("")) {
	    if (names.containsKey(name)) {
	      throw new SyntaxException(
                "Detected two net elements with the same name: "+name+".",null)
                  .addObject(node)
                  .addObject((ShadowNode)names.get(name));
	    } else {
	      names.put(name,node);
	    }
	  }
	}
      }
    }
  }

  private final static String[] lintNames=new String[]{
    "channel check",
    "name check"
  };

  public String[] getLintNames() {
    return lintNames;
  }

  public void lint(int lintNr, ShadowNetSystem netSystem,
		   ShadowNet selectedNet, ShadowNet lastSelectedNet)
    throws SyntaxException
  {
    switch (lintNr) {
    case 0:
      channelCheck(netSystem);
      break;
    case 1:
      doubleNameCheck(netSystem);
      break;
    }
  }
}
