package de.renew.simulator;

import collections.*;
import de.renew.unify.*;
import de.renew.function.Executor;
import de.renew.util.Value;
import de.renew.util.Null;
import java.lang.reflect.Array;

class FlexibleArcBinder implements Binder {
  Variable tokenVar;
  FlexibleArc arc;
  TokenReserver reserver;
  PlaceInstance instance;

  FlexibleArcBinder(Variable tokenVar, TokenReserver reserver,
		    PlaceInstance instance, FlexibleArc arc)
  {
    this.tokenVar=tokenVar;
    this.reserver=reserver;
    this.instance=instance;
    this.arc=arc;
  }

  public int bindingBadness(Searcher searcher) {
    if (Unify.isBound(tokenVar)) {
      return 1;
    } else {
      // We must not try to bind the variable.
      return BindingBadness.max;
    }
  }

  public void bind(Searcher searcher) {
    if (!Unify.isBound(tokenVar)) {
      // This should not happen.
      throw new RuntimeException("Flexible arc binder was invoked for an "+
	"incomplete value.");
    }

    // The array is analysed locally and need not be copied.
    Object tokens=tokenVar.getValue();

    if (tokens==null) {
	// Let's ignore this arc. Probably the null value is meant as
	// a substitute for an empty array.
	searcher.search();
    } else if (tokens.getClass().isArray()) {
	// Make sure that the place instance notifies
	// the searchable if its marking changes,
	// because in that case the possible binding would have
	// to be rechecked. This is only required for input
	// arcs, because output arcs cannot disable or enable
	// a transition.
	searcher.insertTriggerable(instance.triggerables());

	Class elementType=tokens.getClass().getComponentType();

	int n=Array.getLength(tokens);
	
	Object[] tokenValues=new Object[n];
	for (int i=0;i<n;i++) {
	    Object tok=null;
	    try {
		// Extract a single token and copy it, so that
		// backtracking does not hurt.
		tok=Unify.copyBoundValue(Array.get(tokens,i));
	    
		if (elementType.isPrimitive()) {
		    tok=new Value(tok);
		} else {
		    // Is this really neccessary? The array should not
		    // contain any objects that are subject to backtracking.
		    // But let's be safe.
		    tok=Unify.copyBoundValue(tok);
		}

		if (arc.forwardFunction!=null) {
		    Object convertedToken=arc.forwardFunction.function(tok);
	      
		    if (arc.backwardFunction!=null) {
			Object restoredToken=
			    arc.backwardFunction.function(convertedToken);
			Unify.unify(tok,restoredToken,null);
		    }

		    tok=convertedToken;
		}
   
		tokenValues[i]=tok;
	    }
	    catch (Exception e) {
		// Abort this search process.
		// A conversion did not succeed.
		e.printStackTrace();
		return;
	    }
	}

	// Now we start computation with side effects.
	boolean success=true;
	int maxOk=0;
	while (success && maxOk<n) {
	    try {
		success=reserver.removeToken(instance,tokenValues[maxOk],0);
		// Reservation was successful.
		maxOk++;
	    }
	    catch (Exception e) {
		e.printStackTrace();
		success=false;
	    }
	}

	if (success) {
	    // All reservations succeeded.
	    searcher.search();
	}
	  
	// Undo the successful reservations.
	while (maxOk>0) {
	    maxOk--;
	    reserver.unremoveToken(instance,tokenValues[maxOk],0);
	}
    }
  }
}
