package de.renew.simulator;

public class ParallelSimulator implements Simulator {
  public final int multiplicity;
  private final Simulator[] simulators;

  public ParallelSimulator(int multiplicity) {
    if (multiplicity==0) {
      multiplicity=1;
    }
    boolean wantConcurrent=multiplicity>0;
    if (!wantConcurrent) {
      multiplicity=-multiplicity;
    }
    this.multiplicity=multiplicity;

    simulators=new Simulator[multiplicity];
    for (int i=0;i<multiplicity;i++) {
      TracerFactory tracerFactory=new ConsoleTracerFactory(i+"-");
      if (wantConcurrent) {
	simulators[i]=new ConcurrentSimulator(tracerFactory);
      } else {
	simulators[i]=new SequentialSimulator(tracerFactory);
      }
    }
  }

  // In the concurrent simulator, you never know if
  // there might be more enabled transitions. So I am active
  // whenever the creation of the primary instance succeed.
  public boolean isActive() {
    for (int i=0;i<multiplicity;i++) {
      if (simulators[i].isActive()) {
	return true;
      }
    }
    return false;
  }

  // Start the simulation in the background.
  public synchronized void startRun() {
    for (int i=0;i<multiplicity;i++) {
      simulators[i].startRun();
    }
  }

  // Gently stop the simulation.
  public void stopRun() {
    for (int i=0;i<multiplicity;i++) {
      simulators[i].stopRun();
    }
  }

  // Terminate the simulation once and for all.
  // Do some final clean-up and exit all threads.
  public synchronized void terminateRun() {
    for (int i=0;i<multiplicity;i++) {
      simulators[i].terminateRun();
    }
  }

  // Perform just one step or terminate the simulation, if
  // it is running in the background. Return true, if
  // another binding could be found.
  public synchronized int step() {    
    for (int i=1;i<multiplicity;i++) {
      simulators[i].stopRun();
    }
    return simulators[0].step();
  }

  public void refresh() {
    for (int i=1;i<multiplicity;i++) {
      simulators[i].refresh();
    }
  }
}

