package de.renew.simulator;

import de.renew.unify.*;
import collections.*;
import java.io.Serializable;
import java.io.IOException;
import de.renew.util.RenewObjectOutputStream;
import de.renew.event.PlaceEventProducer;
import de.renew.event.PlaceEventListener;
import de.renew.event.TokenEvent;
import de.renew.event.PlaceEvent;

public class Place implements PlaceEventProducer, Serializable {

  static final long serialVersionUID = -454336128139256244L;

  /**
   * My name.
   **/
  String name;

  /**
   * true, if my instances should output a trace message when they
   * receive an initial token. This is the default.
   **/
  private boolean trace;

  /**
   * My set of token source inscriptions.
   **/
  UpdatableSet inscriptions;

  public Place(Net net, String name) {
    this.name=name;

    trace=true;
    inscriptions=new HashedSet();

    net.add(this);
  }

  public String toString() {
    return name;
  }

  PlaceInstance makeInstance(NetInstance netInstance,
			     boolean wantInitialTokens) throws Impossible
  {
    PlaceInstance instance=new PlaceInstance(netInstance,this,
					     wantInitialTokens);
    
    return instance;
  }

  /**
   * Switch my trace flag on or off.
   *
   * @param trace true, if tracing is desired
   **/
  public void setTrace(boolean trace) {
    this.trace=trace;
  }

  /**
   * Am I being traced?
   *
   * @return true, if my trace flag is set
   **/
  public boolean getTrace() {
    return trace;
  }

  /**
   * What's my name?
   *
   * @return my name
   **/
  public String getName() {
    return name;
  }

  /**
   * Give me another token source.
   **/
  public void add(TokenSource tokenSource) {
    inscriptions.include(tokenSource);
  }
  
  /**
   * Remove one my current token sources.
   **/
  public void remove(TokenSource tokenSource) {
    inscriptions.exclude(tokenSource);
  }


  /**
   * Serialization method, behaves like default writeObject
   * method except using the domain trace feature, if the
   * output is a RenewObjectOutputStream.
   * @see de.renew.util.RenewObjectOutputStream
   **/
  private void writeObject(java.io.ObjectOutputStream out) throws IOException {
    RenewObjectOutputStream rOut = null;
    if (out instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)out;
    if (rOut != null) rOut.beginDomain(this);
    out.defaultWriteObject();
    if (rOut != null) rOut.endDomain(this);
  }

  /**
   * Deserialization method, behaves like default readObject
   * method except restoring additional transient fields.
   **/
  private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    listeners = new HashedSet();
  }

  // ---- ID-Handling ------------------------------------------------

  /**
   * Holds the internal ID assigned to this place.
   * @see #setID
   **/
  private int _id;

  /**
   * Assigns the given internal ID (e.g. the ID of the
   * corresponding shadow net element) to this place.
   * This should be done at creation time by the compiler.
   *
   * @see de.renew.shadow.ShadowNetElement#setID
   **/
  public void setID(int newID) {
    _id = newID;
    //System.out.println("Place \""+name+"\" was assigned to ID "+_id);
  }

  /**
   * Returns the internal ID assigned to this place.
   * @see #setID
   **/
  public int getID() {
    return _id;
  }


  // ---- PlaceEvent Handling ----------------------------------------


  /**
   * The listeners to notify if place events occur.
   *
   * This object is used to synchronize listener additions/removals
   * and notification.
   **/
  private transient UpdatableSet listeners=new HashedSet();

  public void addPlaceEventListener(PlaceEventListener listener)
  {
    synchronized(listeners) {
      listeners.include(listener);
    }
  }
  
  public void removePlaceEventListener(PlaceEventListener listener)
  {
    synchronized(listeners) {
      listeners.exclude(listener);
    }
  }

  /**
   * Informs all PlaceEventListeners that an instance
   * has just changed its marking dramatically.
   *
   * This method is called by all instances of this place
   * if they have to send such an event.
   *
   * @param event the event to distribute
   *
   * @see PlaceInstance#markingChanged
   **/
  void markingChanged(PlaceEvent event) {
    synchronized(listeners) {
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.markingChanged(event);
      }    
    }
  }

  /**
   * Informs all PlaceEventListeners that a token
   * was just added to the marking of an instance.
   *
   * This method is called by all instances of this place
   * if they have to send such an event.
   *
   * @param event the event to distribute
   *
   * @see PlaceInstance#tokenAdded
   **/
  void tokenAdded(TokenEvent event) {
    synchronized(listeners) {
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.tokenAdded(event);
      }    
    }
  }
  
  /**
   * Informs all PlaceEventListeners that a token
   * was just removed from the marking of an instance.
   *
   * This method is called by all instances of this place
   * if they have to send such an event.
   *
   * @param event the event to distribute
   *
   * @see PlaceInstance#tokenRemoved
   **/
  void tokenRemoved(TokenEvent event) {
    synchronized(listeners) {
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.tokenRemoved(event);
      }    
    }
  }
  
  /**
   * Informs all PlaceEventListeners that a token
   * was just tested within the marking of an instance.
   *
   * This method is called by all instances of this place
   * if they have to send such an event.
   *
   * @param event the event to distribute
   *
   * @see PlaceInstance#tokenTested
   **/
  void tokenTested(TokenEvent event) {
    synchronized(listeners) {
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.tokenTested(event);
      }    
    }
  }

  /**
   * Informs all PlaceEventListeners that a token
   * was just untested within the marking of an instance.
   *
   * This method is called by all instances of this place
   * if they have to send such an event.
   *
   * @param event the event to distribute
   *
   * @see PlaceInstance#tokenUntested
   **/
  void tokenUntested(TokenEvent event) {
    synchronized(listeners) {
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.tokenUntested(event);
      }
    }    
  }

}

