package de.renew.simulator;

import collections.*;
import java.io.Serializable;
import java.io.IOException;
import de.renew.util.RenewObjectOutputStream;
import de.renew.event.TransitionEventProducer;
import de.renew.event.TransitionEventListener;
import de.renew.event.FiringEvent;

/**
 * I represent a transition with all its semantic information,
 * but without any state. I can have a name, an uplink,
 * and a set of inscriptions. My main job is to store
 * all this information. I do not actually perform any
 * useful services.
 **/
public class Transition implements TransitionEventProducer,  Serializable {

  static final long serialVersionUID = -5900289631849657071L;

  /**
   * My name.
   **/
  private final String name;

  /**
   * true, if my instances should output a trace message when they
   * occur. This is the default.
   **/
  private boolean trace;

  /**
   * My uplink or null, if I am spontaneous.
   **/
  UplinkInscription uplink;
  
  /**
   * A set of inscriptions. I expect all inscription objects to be
   * of the type TransitionInscription.
   *
   * @see de.renew.simulator.TransitionInscription
   **/
  UpdatableSet inscriptions;

  /**
   * I (a transition) am created as a spontaneous transition.
   * Later on, inscriptions may be added, even an uplink.
   * I am automatically registered at the net as a
   * net element.
   *
   * @param net the net to which this transition should be added.
   **/
  public Transition(Net net, String name) {
    this.name=name;

    // Trace actions.
    trace=true;

    // Clear the uplink: no uplink is the default.
    uplink=null;

    // Initialize the array for the inscriptions.
    inscriptions=new HashedSet();

    // Notify the net that we are done and want to be
    // inserted into the list of transitions.
    net.add(this);
  }

  /**
   * I return my name as my string representation.
   * My net instances use my name as a prefix for their name.
   * @return my name.
   **/
  public String toString() {
    return name;
  }

  /**
   * Switch my trace flag on or off.
   *
   * @param trace true, if tracing is desired
   **/
  public void setTrace(boolean trace) {
    this.trace=trace;
  }

  /**
   * Am I being traced?
   *
   * @return true, if my trace flag is set
   **/
  public boolean getTrace() {
    return trace;
  }

  /**
   * Am I spontaneous?
   *
   * @return true, if I am spontaneous, i.e., if I do not
   *   have an uplink
   **/
  public boolean isSpontaneous() {
    return uplink==null;
  }

  /**
   * Here I extract my uplink, if any, from my inscriptions.
   * If there is more than one uplink, it is undetermined which
   * uplink I will use. This method is called after every
   * update to my set of inscriptions.
   **/
  private void checkUplink() {
    // Maintain the count of uplinks and select an uplink
    // from the set of inscriptions. By allowing multiple inscriptions
    // the embedding into a graphical editor becomes easier.
    // On the other hand, the transition might get into
    // an inconsistent state. We try to make it as simulatable as
    // possible, however.
    uplink=null;
    CollectionEnumeration enum=inscriptions.elements();
    while (enum.hasMoreElements()) {
      Object inscription=enum.nextElement();
      if (inscription instanceof UplinkInscription) {
	uplink=(UplinkInscription)inscription;
      }
    }
  }

  /**
   * Give me another inscription. After inserting
   * the inscriptions, I will recheck my uplink.
   **/
  public void add(TransitionInscription transitionInscription) {
    if (!inscriptions.includes(transitionInscription)) {
      inscriptions.include(transitionInscription);
      checkUplink();
    }
  }
  
  /**
   * Remove one my current inscriptions. Check if an
   * uplink is still provided.
   **/
  public void remove(TransitionInscription transitionInscription) {
    if (inscriptions.includes(transitionInscription)) {
      inscriptions.exclude(transitionInscription);
      checkUplink();
    }
  }


  /**
   * Serialization method, behaves like default writeObject
   * method except using the domain trace feature, if the
   * output is a RenewObjectOutputStream.
   * @see de.renew.util.RenewObjectOutputStream
   **/
  private void writeObject(java.io.ObjectOutputStream out) throws IOException {
    RenewObjectOutputStream rOut = null;
    if (out instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)out;
    if (rOut != null) rOut.beginDomain(this);
    out.defaultWriteObject();
    if (rOut != null) rOut.endDomain(this);
  }

  /**
   * Deserialization method, behaves like default readObject
   * method except restoring additional transient fields.
   **/
  private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    listeners = new HashedSet();
  }

  // ---- ID-Handling ------------------------------------------------

  /**
   * Holds the internal ID assigned to this transition.
   * @see #setID
   **/
  private int _id;

  /**
   * Assigns the given internal ID (e.g. the ID of the
   * corresponding shadow net element) to this transition.
   * This should be done at creation time by the compiler.
   *
   * @see de.renew.shadow.ShadowNetElement#setID
   **/
  public void setID(int newID) {
    _id = newID;
    //System.out.println("Transition \""+name+"\" was assigned to ID "+_id);
  }

  /**
   * Returns the internal ID assigned to this transition.
   * @see #setID
   **/
  public int getID() {
    return _id;
  }

  // ---- Transition firing event handling ---------------------------

  /**
   * The listeners that I must notify.
   **/
  private transient UpdatableSet listeners=new HashedSet();

  public synchronized void addTransitionEventListener(
                           TransitionEventListener listener)
  {
    listeners.include(listener);
  }

  public synchronized void removeTransitionEventListener(
                           TransitionEventListener listener)
  {
    listeners.exclude(listener);
  }

  /**
   * Informs all TransitionEventListeners that an instance
   * has just started firing.
   *
   * This method is called by all instances of this transition
   * if they have to send such an event.
   *
   * @see TransitionInstance#firingStarted
   **/
  synchronized void firingStarted(FiringEvent fe) {
    CollectionEnumeration enum=listeners.elements();
    while (enum.hasMoreElements()) {
      TransitionEventListener listener=(TransitionEventListener)enum.nextElement();
      listener.firingStarted(fe);
    }    
  }

  /**
   * Informs all TransitionEventListeners that an instance
   * has just completed firing.
   *
   * This method is called by all instances of this transition
   * if they have to send such an event.
   *
   * @see TransitionInstance#firingComplete
   **/
  synchronized void firingComplete(FiringEvent fe) {
    CollectionEnumeration enum=listeners.elements();
    while (enum.hasMoreElements()) {
      TransitionEventListener listener=(TransitionEventListener)enum.nextElement();
      listener.firingComplete(fe);
    }    
  }

}

