package de.renew.unify;

import collections.*;

// On the fly:
//   Check that no calculated unknown is bound.
//   Check that no calculated unknown is registered twice.
//   Check that no two calculated unknowns depend on each other.
// During the consistency check:
//   Check that all early variables have become bound.
//   Check that all late variables are either bound or calculated.

final public class CalculationChecker {
  private UpdatableSet lateVariables;
  private UpdatableSet earlyVariables;

  public CalculationChecker() {
    lateVariables=new HashedSet();
    earlyVariables=new HashedSet();
  }

  public void reset() {
    lateVariables.clear();
    earlyVariables.clear();
  }

  public void addEarlyVariable(final Variable var, StateRecorder recorder)
    throws Impossible 
  {
    if (!earlyVariables.includes(var)) {
      if (recorder!=null) {
	recorder.record(new StateRestorer() {
	  public void restore() {
	    earlyVariables.exclude(var);
	  }
	});
      }
      
      earlyVariables.include(var);
    }
  }

  public void addCalculated(Class targetType, Object target, Object source,
    StateRecorder recorder) throws Impossible 
  {
    Unify.unify(target,new Calculator(targetType,source,recorder),recorder);
  }

  public void addLateVariable(final Variable var, StateRecorder recorder)
    throws Impossible  
  {
    if (!lateVariables.includes(var)) {
      if (recorder!=null) {
	recorder.record(new StateRestorer() {
	  public void restore() {
	    lateVariables.exclude(var);
	  }
	});
      }
      
      lateVariables.include(var);
    }
  }
  
  private boolean checkLateVariables() {
    // All late variables must be bound or calculated.
    CollectionEnumeration enum=lateVariables.elements();
    while (enum.hasMoreElements()) {
      Variable variable=(Variable)enum.nextElement();
      if (!variable.isComplete()) {
	return false;
      }
    }
    return true;
  }

  private boolean checkEarlyVariables() {
    // All early variables must be bound.
    CollectionEnumeration enum=earlyVariables.elements();
    while (enum.hasMoreElements()) {
      Variable variable=(Variable)enum.nextElement();
      if (!variable.isBound()) {
	return false;
      }
    }
    return true;
  }

  public boolean isConsistent() {
    return checkLateVariables() && checkEarlyVariables();
  }
}
