package de.uni_hamburg.fs;

import collections.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.NoSuchElementException;

public class ConceptImpl extends SimpleConcept {

  private String name = null;
  private boolean newExtensional = false, extensional = false;
  private UpdatableSet subsumes, subsumedBy;
  protected ConceptSet directSubs = new ConceptSet(),
      directSupers = new ConceptSet();
  private OrderedTable newApprop = new OrderedTable(), approp=null;
  private boolean restricted = true;
  UpdatableSet partitions=new HashedSet(); // of Partitions

  ConceptImpl() {
    name = "Top";
    transitiveClosure();
  }

  public ConceptImpl(String name, ConceptImpl[] directSuperList)
      throws TypeException {
    this(name);
    for (int i = 0; i < directSuperList.length; ++i)
      addIsa(directSuperList[i]);
  }

  public ConceptImpl(String name, ConceptImpl superConcept)
      throws TypeException{
    this(name, new ConceptImpl[] {superConcept});
  }

  public ConceptImpl(String name) {
      subsumes = new HashedSet();
      subsumes.include(this);
      if (!setName(name))
	  throw new RuntimeException("Concept "+name+" already exists!");
      TypeSystem.instance().transitiveClosure();
  }

  public ConceptImpl(String name, boolean extensional, boolean restricted) {
     this(name);
     this.extensional = extensional;
     this.restricted = restricted;
  }

  /** Return the name of this Concept. */
  public String getName() { return name; }

  /** Change the name of this Concept.
    * <p>A concept must have a unique name! If the new name already
    * exists, nothing is changed.
    */
  public boolean setName(String newname) {
    if (newname == null || newname.length() == 0)
      return true;
    if (newname.equals(name))
      return true;
    TypeSystem ts=TypeSystem.instance();
    if (ts.hasConcept(newname)) {
      System.err.println("Concept "+newname+" already exists!");
      return false;
    }
    if (name != null)
      ts.removeConcept(this);
    name = newname;
    ts.addConcept(this);
    return true;
  }

  /** Return whether this Concept is extensional. */
  public boolean isExtensional() {
    return extensional;
  }

  /** Return whether this Concept restricts which features are allowed.
   *  A non-restricted Concept is appropriate for every feature.
   */
  public boolean isRestricted() {
      return restricted;
  }

  /** Return whether the feature <feature> is appropriate in this Concept. 
   */
  public boolean isApprop(Name feature) {
      //if (approp==null)
      //    System.out.println("Approp is null in "+name+" ("+hashCode()+") !!!");

    return !restricted || approp.includesKey(feature);
  }

  /** Return the required Type of the Value found under the given feature. */
  public Type appropType(Name featureName) throws NoSuchElementException {
      try {
	  return new ConjunctiveType(appropConcepts(featureName));
      } catch (UnificationFailure uff) {
	  throw new RuntimeException("The ConjunctiveType "+appropConcepts(featureName)+" is not defined!");
      }
  }

  public ConceptSet appropConcepts(Name featureName) throws NoSuchElementException {
    if (!restricted) {
	return new ConceptSet();
    }
    if (approp.includesKey(featureName)) {
	return (ConceptSet)approp.at(featureName);
    }
    throw new NoSuchElementException();
  }


  /** Return an Enumeration of all appropriate features. */
  public Enumeration appropFeatureNames() {
      if (restricted)
	  return approp.keys();
      else
	  return new Vector().elements();
  }

  void inherit(UpdatableSet visited) throws TypeException {
    visited.include(this);
    ConceptEnumeration subenum = directSubs.elements();
    while (subenum.hasMoreElements()) {
      ConceptImpl sub = (ConceptImpl)subenum.nextConcept();
      if (extensional) {
	if (!sub.extensional) {
	  sub.extensional = true;
	  sub.subsumedBy = new HashedSet();
	  sub.subsumedBy.include(sub);
	}
	sub.subsumedBy.include(this);
      }
  // System.err.println("Concept: "+name+" approp: "+approp);
      boolean subchanged=false;
      int subcount=0;
      Enumeration appropenum = approp.keys();
      while (appropenum.hasMoreElements()) {
	Name featureName = (Name)appropenum.nextElement();
	ConceptSet appropType = (ConceptSet)approp.at(featureName);
	if (sub.isApprop(featureName)) {
	  ConceptSet subapprop=(ConceptSet)sub.approp.at(featureName);
	  appropType.unite(subapprop);
	  sub.approp.insertAt(featureName,subcount++,appropType);
	  if (!appropType.equals(subapprop)) {
	      subchanged = true;
	  }
	} else {
	  sub.approp.insertAt(featureName,subcount++,appropType);
	  subchanged = true;
	}
      }
      if (subchanged || !visited.includes(sub)) {
	sub.inherit(visited);
      }
    }
  }

  protected void transitiveClosure() {
    subsumes = new HashedSet();
    subsumes.include(this);
    ConceptEnumeration subenum = directSubs.elements();
    while (subenum.hasMoreElements()) {
      ConceptImpl sub = (ConceptImpl)subenum.nextConcept();
      sub.transitiveClosure();
      subsumes.includeElements(sub.subsumes.elements());
    }
  }

  void recalcDirectIsa(UpdatableSet visited) throws TypeException {
      if (!visited.includes(this)) {
	  visited.include(this);
	  try {
	      directSupers.join();
	  } catch (UnificationFailure uff) {
	      TypeException tex=new TypeException(this);
	      throw tex;
	  }
	  directSubs.meet();

	  extensional = newExtensional;
	  if (extensional) {
	      subsumedBy = new HashedSet();
	      subsumedBy.include(this);
	  } else {
	      subsumedBy = null;
	  }
	  approp=(OrderedTable)newApprop.clone();

	  Enumeration subenum = directSubs.elements();
	  while (subenum.hasMoreElements()) {
	      ((ConceptImpl)subenum.nextElement()).recalcDirectIsa(visited);
	  }
      }
  }

  void basicAddIsa(ConceptImpl superconcept) {
    directSupers.addConcept(superconcept);
    superconcept.directSubs.addConcept(this);
    //System.out.println("Adding "+this+" isa "+superconcept);
  }

  public void addIsa(ConceptImpl superconcept)
      throws CyclicHierarchyException,
	     TypeException {
    if (superconcept.isa(this))
      throw new CyclicHierarchyException();
    if (isa(superconcept))
      return;
    basicAddIsa(superconcept);
    TypeSystem.instance().recalcDirectIsa();
  }

  public void addApprop(String feature, ConceptSet type) {
    addApprop(new Name(feature),type);
  }

  public void addApprop(Name featureName, ConceptSet type) {
    //System.out.println("Feature "+featureName+" defined in "+getName());
    newApprop.putAt(featureName,type);
  }


  public void setApprops(OrderedTable approps) {
    newApprop = (OrderedTable)approps.clone();
  }

  void buildFeatureTypes(UpdatableSet visited) throws TypeException {
      if (!visited.includes(this)) {
	  visited.include(this);
	  //System.out.println("Converting feature types for concept "+this);
	  Enumeration featenum=approp.keys();
	  while (featenum.hasMoreElements()) {
	      Name feature=(Name)featenum.nextElement();
	      //System.out.println("Converting feature "+feature);
	      ConceptSet cs=(ConceptSet)approp.at(feature);
	      try {
		  new ConjunctiveType(cs);
	      } catch (UnificationFailure uff) {
		  throw new TypeException(this,feature);
	      }
	  }
	  Enumeration subenum = directSubs.elements();
	  while (subenum.hasMoreElements()) {
	      ((ConceptImpl)subenum.nextElement()).buildFeatureTypes(visited);
	  }
      }
  }

  /** Return whether this Concept is-a <that> Concept.
    * In other words, return wheter this Concept is more special than <that>
    * Concept.
    */
  public boolean isa(Concept that) {
     if (that instanceof ConceptImpl) {
        return ((ConceptImpl)that).subsumes.includes(this);
     } else {
        return false;
     }
  }

  /** Return the Partition this ConceptImpl belongs to. */
  public Enumeration getPartitions() {
      return partitions.elements();
  }

  /** Return whether this Concept is-not-a <that> Concept.
    * In other words, return wheter this Concept is incompatible with <that>
    * Concept.
    */
  public boolean isNotA(Concept that) {
     if (this == that)
        return false;
     if (that instanceof ConceptImpl) {
        ConceptImpl thatCon = (ConceptImpl)that;
        if (intersects(partitions,thatCon.partitions))
            return true;
	return superIsNotA(thatCon) || thatCon.superIsNotA(this);
	// TODO: this isNotA that is also true if some appropType
	// of this is incompatible with the corresponding appropType
	// of that. Caution: recursive definition!!!
     }
     return true;
  }

  private boolean intersects(Set a, Set b) {
      if (a.size() > b.size())
	  return intersects(b,a);
      Enumeration as=a.elements();
      while (as.hasMoreElements()) {
	  if (b.includes(as.nextElement()))
	      return true;
      }
      return false;
  }

  private boolean superIsNotA(ConceptImpl that) {
    ConceptEnumeration superenum = directSupers.elements();
    while (superenum.hasMoreElements()) {
       Concept superCon = superenum.nextConcept();
       if (superCon.isNotA(that))
          return true;
    }
    return false;
  }

  public ConceptEnumeration extensionalSuperconcepts() {
    return new ConceptEnumeration(subsumedBy.elements());
  }

    /*
  public String toString() {
    StringBuffer output = new StringBuffer(name);
    if (extensional)
      output.append(" (extensional)");

    //if (approp==null)
    //    System.out.println("Approp is null in "+name+" ("+hashCode()+") !!!");

    Enumeration appropenum = approp.keys();
    while (appropenum.hasMoreElements()) {
      Name featureName = (Name)appropenum.nextElement();
      Type appropType = (Type)approp.at(featureName);
      output.append("\n ").append(featureName.toString())
	.append(":").append(appropType.getName());
    }
    output.append("\nDirect Supers: ").append(directSupers.toString());
    output.append("\nDirect Subs: ").append(directSubs.toString());
    return output.toString();
  }
    */

}
