package de.uni_hamburg.fs;

import collections.*;
import java.util.Enumeration;
import java.util.StringTokenizer;

public class ConjunctiveType implements Type {

  private UpdatableSet concepts = new HashedSet();

  // "caches" for aggregated concept attributes:
  private OrderedTable features = new OrderedTable();
  private boolean extensional=false;
  private boolean restricted=true;

  protected int hashCode; // as the hashCode is a finicky computation, it is being cached

  ConjunctiveType () {
     hashCode = 0;
  }

  public ConjunctiveType(Concept concept) {
     try {
	 addConcept(concept);
     } catch (UnificationFailure uff) {
	 // Should not happen.
	 throw new RuntimeException("***Unexpected failure in addConcept("+concept+").");
     }
  }

  public ConjunctiveType(ConceptSet concepts) throws UnificationFailure {
      ConceptEnumeration conceptEnum=concepts.elements();
      while (conceptEnum.hasMoreElements()) {
	  addConcept(conceptEnum.nextConcept());
      } 
  }

  public int hashCode() { return hashCode; }

  /** Return an Enumeration of all Concepts this ConjunctiveType is composed 
   *  of.
   */
  public CollectionEnumeration concepts() {
    return concepts.elements();
  }

  /** Return the name of this Type. */
  public String getName() {
    StringBuffer output = new StringBuffer();
    Enumeration conceptEnum = concepts.elements();
    for (int i = 1; conceptEnum.hasMoreElements(); ++i) {
      if (i > 1)
	output.append(',');
      Concept concept = (Concept)conceptEnum.nextElement();
      output.append(concept.getName());
    }
    String name=output.toString();
    if (name.length()==0)
	return "Top";
    else
	return name;
  }

  /** Finds or constructs a ConjunctiveType consisting of the given
    * Concepts. This may fail due to incompatibilities.
    */
  public static Type getType(Set concepts) throws UnificationFailure {
    if (concepts.isEmpty())
       return Type.TOP;
    Enumeration conceptEnum = concepts.elements();
    ConjunctiveType  newtype = new ConjunctiveType();
    while (conceptEnum.hasMoreElements()) {
      newtype.addConcept((Concept)conceptEnum.nextElement());
    }
    return newtype;
   }

  public static Type getType(String conceptsStr) 
    throws UnificationFailure, TypeException {
    StringTokenizer conceptNames = new StringTokenizer(conceptsStr,",");
    HashedSet concepts = new HashedSet();
    TypeSystem ts = TypeSystem.instance();
    while (conceptNames.hasMoreElements()) {
        concepts.include(ts.conceptForName(conceptNames.nextToken()));
    }
    return getType(concepts);
  }

  public boolean isExtensional() {
    return extensional;
  }

  public boolean isApprop(Name featureName) {
    if (restricted)
	return features.includesKey(featureName);
    else
	return true;
  }

  public Type appropType(Name featureName) {
    try {
      if (restricted || features.includesKey(featureName)) {
	  return new ConjunctiveType((ConceptSet)features.at(featureName));
      } else {
	  return Type.TOP;
      }
    } catch (UnificationFailure uff) {
	throw new RuntimeException("***Conjunctive type for ConceptSet "+features.at(featureName)+" not defined!");
    }
  }

  public Enumeration appropFeatureNames() {
    return features.keys();
  }

  public boolean subsumes(Type that) {
    Enumeration conceptEnum = concepts.elements();
  forAllConcepts:
    while (conceptEnum.hasMoreElements()) {
       Concept concept = (Concept)conceptEnum.nextElement();
       Enumeration thatConceptEnum = that.concepts();
       while (thatConceptEnum.hasMoreElements()) {
           Concept thatConcept = (Concept)thatConceptEnum.nextElement();
	   if (thatConcept.isa(concept))
                continue forAllConcepts;
        }
        return false;        // No that-concept found!
    }    
    return true;
  }


  /** Return whether this Type is subsumed by <that> Type.
    * In other words, return wheter this Type is more special than <that>
    * Type.
    */
  public boolean subsumedBy(Type that) {
     return that.subsumes(this);
  }

  public boolean canUnify(Type that) {
   // TODO: canUnify also has to unify Types of redefined features!
   Enumeration conceptEnum = concepts.elements();
    while (conceptEnum.hasMoreElements()) {
       Concept concept = (Concept)conceptEnum.nextElement();
       Enumeration thatConceptEnum = that.concepts();
       while (thatConceptEnum.hasMoreElements()) {
           Concept thatConcept = (Concept)thatConceptEnum.nextElement();
	   if (thatConcept.isNotA(concept))
                return false;
        }
    }    
    return true;
   }

   private void basicAddConcept(Concept newConcept) throws UnificationFailure {
       Enumeration featenum=newConcept.appropFeatureNames();
       while (featenum.hasMoreElements()) {
	   Name feature=(Name)featenum.nextElement();
	   ConceptSet newType=newConcept.appropConcepts(feature);
	   if (features.includesKey(feature)) {
	       // this may fail:
	       newType.joinConcepts(((ConceptSet)features.at(feature)).elements());
	   }
	   features.putAt(feature,newType);
       }
       extensional = extensional || newConcept.isExtensional();
       restricted = restricted && newConcept.isRestricted();
   }

  /** Adds a part Type to this ConjunctiveType. This method should
    * only be used during the construction of a new Type.
    */
  public void addConcept(Concept newConcept) throws UnificationFailure {
      basicAddConcept(newConcept); // add features & extensionality of newConcept
      if (concepts.isEmpty()) {
	  concepts.include(newConcept);
	  hashCode += newConcept.hashCode();
	  return;
      }
      UpdatableSet newConcepts = new HashedSet();
      Enumeration conceptEnum = concepts();
      while (conceptEnum.hasMoreElements()) {
        Concept concept = (Concept)conceptEnum.nextElement();
	ConceptSet uni=concept.unify(newConcept);
	newConcepts.includeElements(uni.elements());
      }
      concepts = newConcepts;
      hashCode = 0;
      conceptEnum = concepts();
      while (conceptEnum.hasMoreElements()) {
        hashCode += conceptEnum.nextElement().hashCode();
      }
  }

  private ConjunctiveType duplicate() {
      ConjunctiveType  typ = new ConjunctiveType ();
      typ.concepts = (UpdatableSet)concepts.duplicate();
      typ.features = (OrderedTable)features.clone();
      typ.extensional = extensional;
      typ.restricted = restricted;
      typ.hashCode = hashCode;
      return typ;
  }

  public Type unify(Type that) throws UnificationFailure {
      ConjunctiveType  typ = duplicate();
      Enumeration thatConceptEnum = that.concepts();
      while (thatConceptEnum.hasMoreElements()) {
          Concept thatConcept = (Concept)thatConceptEnum.nextElement();
          typ.addConcept(thatConcept);
      }
      return typ;
  }

  public Type mostGeneralExtensionalSupertype(Type that) {
     return null;
/*
//     System.out.println("Finding most general extensional supertype for "+this+" and "+that+".");
    UpdatableSet thisSupers = new HashedSet();
    Enumeration conceptEnum = concepts.elements();
    while (conceptEnum.hasMoreElements()) {
      Type type = (Concept)conceptEnum.nextElement();
      thisSupers.includeElements(type.extensionalSuperconcepts());
    }
//    System.out.println("Supertype of "+this+": "+new ConjunctiveType(thisSupers));
    UpdatableSet commonSupers = new HashedSet();
    Enumeration thatConcs = that.concepts();
    while (thatConcs.hasMoreElements()) {
      Type thattype = (Concept)thatConcs.nextElement();
      Enumeration thatSupers = thattype.extensionalSuperconcepts();
      while (thatSupers.hasMoreElements()) {
         Type thatSuper = (Type)thatSupers.nextElement();
//    System.out.print("Supertype of "+that+" "+thatSuper.getName());
         if (thisSupers.includes(thatSuper)) {
//            System.out.println(" is added!");
            if (!(new ConjunctiveType(commonSupers).subsumes(new ConjunctiveType(thatSuper)))) {
               UpdatableSet newCS = new HashedSet();
//               System.out.print("...and removes some more special concepts:");
               Enumeration oldCSenum = commonSupers.elements();
               while (oldCSenum.hasMoreElements()) {
                 Type oldSuper = (Type)oldCSenum.nextElement();
                 if (!thatSuper.subsumes(oldSuper)) {
                    newCS.include(oldSuper);
                 }
//                 else
//                    System.out.print(" "+oldSuper.getName());
               }
               commonSupers = newCS;
//               System.out.println();
            }
            commonSupers.include(thatSuper);
         }
//         else
//           System.out.println(" ruled out.");
      }
    }
//    System.out.println("Most general extensional supertype for "+this+" and "+that+" is "+new ConjunctiveType(commonSupers)+".");
    return new ConjunctiveType(commonSupers);
*/
  }

  public String toString() {
      return getName();
  }

   /** Compares this ConjunctiveType with another Object. Two ConjunctiveTypes
    * are considered equal iff they consist of the same set of Concepts.
    */ 
 public boolean equals(Object that) {
     if (that instanceof ConjunctiveType) {
        return equalSet(concepts,((ConjunctiveType)that).concepts);
     } else {
        return false;
     }
   }

  private static boolean equalSet(Set a, Set b) {
     if (a.size() != b.size())
        return false;
     Enumeration aEnum=a.elements();
     for (int i=a.size(); i>0; --i) {
        if (!b.includes(aEnum.nextElement()))
          return false;
     }
     return true;
  }

}
