package de.uni_hamburg.fs;

import collections.*;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.NoSuchElementException;

public class FSNode implements Node {
  Type nodetype;
  private OrderedTable feats;

  private static int counter=0;
  public final int seqNo=++counter;

  public FSNode(Type nodetype, Map feats) throws TypeException {
    this.nodetype = nodetype;
    this.feats = new OrderedTable(nodetype.appropFeatureNames());
    Enumeration features = feats.keys();
    while (features.hasMoreElements()) {
      Name feature = (Name)features.nextElement();
      addFeature(feature,(Node)feats.at(feature));
    }
  }

  public FSNode(Type nodetype) {
    this.nodetype = nodetype;
    this.feats = new OrderedTable(nodetype.appropFeatureNames());
  }

  public FSNode() {
    this(TypeSystem.any());
  }

//    public FSNode(Feature feat) throws TypeException { this(new Feature[]{feat}); }

//    public FSNode(Type nodetype, Feature feat) throws TypeException {
//      this(nodetype, new Feature[]{feat});
//    }

//    public FSNode(Feature[] features) throws TypeException {
//      this(Type.TOP, features);
//    }

//    public FSNode(Type nodetype, Feature[] features) throws TypeException {
//       this(nodetype);
//       for (int i=0; i < features.length; ++i)
//          addFeature(features[i].feature, features[i].value);
//    }

  public FSNode(String nodetype) throws UnificationFailure {
    this(ConjunctiveType.getType(nodetype));
  }

  public FSNode(String nodetype, OrderedTable feats)
      throws UnificationFailure, TypeException {
    this(ConjunctiveType.getType(nodetype),feats);
  }

  FSNode(Type nodetype, OrderedTable feats) {
    this.nodetype = nodetype;
    this.feats =  feats;
  }

  public Type getType() { return nodetype; }

  public void refine(Type type) throws UnificationFailure {
     nodetype = nodetype.unify(type);
  }

  public Enumeration featureNames() {
      return feats.keys();
  }

  public boolean hasFeature(Name featureName) {
     return feats.includesKey(featureName);
  }

  public Node delta(Name featureName) throws NoSuchElementException {
    return (Node)feats.at(featureName);
  }

  /** Returns the Node at the given path.
    *  A path is a sequence of Features (meant to be starting at the root node).
    *  The syntax of a path is feature_1:feature_2:...feature_n.
    *  The empty String returns the Node itself.
    *  The exception is thrown if at any point, the feature given
    *  by the path does not exist in the current Node.
    */
  public Node at(String path) throws NoSuchElementException {
     StringTokenizer pathenum=new StringTokenizer(path,":");
     Node curr=this;
     while (pathenum.hasMoreTokens()) {
        curr=curr.delta(new Name(pathenum.nextToken()));
     }
     return curr;
  }

  /** Updates the value of an existing feature.
    *  This method should only be called during construction of
    *  a FSNode and should not be used to create new features.
    *  Use addFeature() instead!
    */
  public void updateFeature(Name featureName, Node newValue) {
     if (!hasFeature(featureName)) {
       throw new RuntimeException("Feature "+featureName+" updated but non-existent!");
     }
     feats.putAt(featureName, newValue);
  }

  /** Adds a new feature with the given name and value.
    *  This method should only be called during construction of
    *  a FSNode and should not be used to change existing features.
    *  Use updateFeature() instead!
    */
  public FSNode addFeature(String feature, Node value) throws TypeException {
    return addFeature(new Name(feature),value);
  }

  /** Adds a new feature with the given name (as feature code) and value.
    *  This method should only be called during construction of
    *  a FSNode and should not be used to change existing features.
    *  Use updateFeature() instead!
    */
  public FSNode addFeature(Name featureName, Node value) throws TypeException {
     if (hasFeature(featureName)) {
       throw new RuntimeException("Feature "+featureName.toString()+" redefined!");
     }
     if (nodetype.isApprop(featureName) && nodetype.appropType(featureName).subsumes(value.getType())) {
       feats.putAt(featureName, value);
       return this;
     } else {
	System.err.println("Ou-Ha! feature: "+featureName+", isApprop: "+nodetype.isApprop(featureName)+", appropType: "+nodetype.appropType(featureName)+", nodetype: "+nodetype+" (+"+nodetype.hashCode()+"), class: "+nodetype.getClass());
        throw new TypeException();
     }
   }

  /** Add all missing features using the most general appropriate
   *  Type for their value node.
   */
  public void totallyWellType() {
      Enumeration featenum=nodetype.appropFeatureNames();
      while (featenum.hasMoreElements()) {
	  Name feat=(Name)featenum.nextElement();
	  if (!hasFeature(feat)) {
	      feats.putAt(feat,new FSNode(nodetype.appropType(feat)));
	  }
      }
  }


  public Node duplicate() {
     return new FSNode(nodetype, (OrderedTable)feats.clone());
  }

}
