package de.uni_hamburg.fs;

import collections.*;
import java.util.Enumeration;

public class TagMap {

   private UpdatableMap tagmap=new HashedMap();
   private UpdatableSet visited=new HashedSet();
   private Node root;

   /** Builds a tag map for a Feature Structure.
      * The tag map provides an Integer for every node that is connected
      * with the given root node. If a node is labelled zero, it has exactly
      * one predescessor and need not be tagged. All other nodes have
      * multiple predecessors and get assigned a unique Integer that can
      * be used as a tag number.
      */
   public TagMap(Node node) {
        root=node;
        buildTagMap(root,0);
   }

    /** Makes a TagMap from a given root node and a Map of tags
     *  that was e.g. the result of parsing a feature structure.
     */
   public TagMap(Node root, EquivRelation er, Map parsedTags) {
       this.root=root;
       Enumeration ptenum=parsedTags.keys();
       while (ptenum.hasMoreElements()) {
	   Integer tag=(Integer)ptenum.nextElement();
	   Node node=er.getUnificator((Node)parsedTags.at(tag));
           tagmap.putAt(node,tag);
       }
   }

   public Node getRoot() { return root; }

   private int buildTagMap(Node node, int tagcounter) {
    int tag;
    if (tagmap.includesKey(node)) {
      tag = ((Integer)tagmap.at(node)).intValue();
      if (tag == 0)     // visited before
	tagmap.putAt(node,new Integer(++tagcounter)); // give node a name
    } else {
      tagmap.putAt(node,new Integer(0)); // mark as visited
      // do this recursively for all nodes:
      Enumeration features = node.featureNames();
      while (features.hasMoreElements()) {
	Name featureName = (Name)features.nextElement();
	tagcounter = buildTagMap(node.delta(featureName), tagcounter);
      }
    }
    return tagcounter;
  }

  public void resetVisited() {
     visited=new HashedSet();
  }

  public final static String indent(int depth) {
    StringBuffer space = new StringBuffer(depth+1);
    for (int i = 0; i < depth; ++i)
      space.append(' ');
    return space.toString();
  }

  public int getTag(Node node) {
    if (tagmap.includesKey(node))
      return ((Integer)tagmap.at(node)).intValue();
   else
      return 0;
  }

  public boolean isVisited(Node node) {
    return visited.includes(node);
  }

  public void setVisited(Node node) {
     visited.include(node);
  }

}
