package de.uni_hamburg.fs;

/** A TypeSystem consists of a set of user-defined concepts toghether
 *  with their inheritance relation. 
 *  There are some predefined concepts available in all TypeSystems,
 *  like the TOPCONCEPT, ANY, and LIST concepts.
 *  Also, all Java classes which are reachable through the CLASSPATH
 *  are available as JavaConcepts.
 */

import collections.*;
import java.util.NoSuchElementException;

public class TypeSystem {


  public static final Name ANY = new Name("Any");
  public static final Name LIST = new Name("List");
  public static final Name NELIST = new Name("NEList");
  public static final Name ELIST = new Name("EList");
  public static final Name HEAD_FEATURE = new Name("hd");
  public static final Name TAIL_FEATURE = new Name("tl");

  private Type tObject;
  private Type tAny;
  private Type tList;
  private Type tEList;
  private Type tNEList;

  private static String imported[] = {"","java.lang."};

  private static TypeSystem _instance=null;

  private ConceptImpl root = new ConceptImpl();
  private UpdatableMap concepts = new HashedMap();

  public TypeSystem() {
    System.out.println("New TypeSystem.");
  }

  private void init() {
    getJavaConcept(Boolean.TYPE);
    getJavaConcept(Byte.TYPE);
    getJavaConcept(Character.TYPE);
    getJavaConcept(Short.TYPE);
    getJavaConcept(Integer.TYPE);
    getJavaConcept(Long.TYPE);
    getJavaConcept(Float.TYPE);
    getJavaConcept(Double.TYPE);

   try {
       tObject = new ConjunctiveType(conceptForName("java.lang.Object"));
       ConceptImpl any = new ConceptImpl(ANY.toString(), false, false);
       ConceptImpl list = new ConceptImpl(LIST.toString(), false, true);
       ConceptImpl elist = new ConceptImpl(ELIST.toString(), true, true);
       elist.basicAddIsa(list);
       ConceptImpl nelist = new ConceptImpl(NELIST.toString(), false, true);
       nelist.addApprop(HEAD_FEATURE,new ConceptSet());
       nelist.addApprop(TAIL_FEATURE,new ConceptSet(list));
       nelist.basicAddIsa(list);
       // ELIST and NELIST are in the same partition:
       new Partition(elist).addConcept(nelist);
    } catch (TypeException tee) {
       // should never happen:
       System.err.println("Exception during TypeSystem init:");
       tee.printStackTrace();
    }
    transitiveClosure();
  }

  public void transitiveClosure() {
      root.transitiveClosure();
  }

  public void recalcDirectIsa() throws TypeException {
      root.transitiveClosure();
      root.recalcDirectIsa(new HashedSet());
  }


  public void inheritFeatures() throws TypeException {
    recalcDirectIsa();
    root.inherit(new HashedSet());
    root.buildFeatureTypes(new HashedSet());
    tAny = new ConjunctiveType(conceptForName(ANY));
    tList = new ConjunctiveType(conceptForName(LIST));
    tEList = new ConjunctiveType(conceptForName(ELIST));
    tNEList = new ConjunctiveType(conceptForName(NELIST));
  }

  public Type tObject() { return tObject; }
  public Type tAny() { return tAny; }
  public Type tList() { return tList; }
  public Type tEList() { return tEList; }
  public Type tNEList() { return tNEList; }


  public static TypeSystem instance() {
      if (_instance==null)
	  return newInstance();
      else
	  return _instance;
  }

  public static TypeSystem newInstance() {
      _instance=new TypeSystem();
      _instance.init();
      return _instance;
  }

  public static void setInstance(TypeSystem instance) {
      _instance=instance;
  }

  public static Type object() { return _instance.tObject; }
  public static Type any() { return _instance.tAny; }
  public static Type list() { return _instance.tList; }
  public static Type eList() { return _instance.tEList; }
  public static Type neList() { return _instance.tNEList; }


  public void addConcept(Concept concept) {
      if (concept instanceof ConceptImpl)
	  ((ConceptImpl)concept).basicAddIsa(root);
      concepts.putAt(new Name(concept.getName()),concept);
  }

  public void removeConcept(Concept concept) {
      concepts.removeAt(concept.getName());
  }

  public boolean hasConcept(String name) {
      return hasConcept(new Name(name));
  }

  public boolean hasConcept(Name name) {
      if (concepts.includesKey(name)) {
	  return true;
      }
      try {
	  getJavaClass(name.toString());
	  return true;
      } catch (TypeException tee) {
	  return false;
      }
  }

  public Concept conceptForName(String name) {
      return conceptForName(new Name(name));
  }

  public Concept conceptForName(Name name) {
      Concept concept=null;
      try {
	  concept = (Concept)concepts.at(name);
      } catch (NoSuchElementException nse) {
	  try {
	      concept=getJavaConcept(getJavaClass(name.toString()));
	  } catch (TypeException tee) {
	      System.out.println("Automatically creating new concept "+name);
	      concept=new ConceptImpl(name.toString(),false,true);
	  }
      }
      return concept;
  }

  public JavaConcept getJavaConcept(Class javaClass) {
    String name=javaClass.getName();
    if (concepts.includesKey(name))
       return (JavaConcept)concepts.at(name);
    else {
       JavaConcept jc = new JavaConcept(javaClass);
       addConcept(jc);
       return jc;
    }
  }

  public Class getJavaClass(String javaClassName)
           throws TypeException {
      int endIndex = imported.length;
      if (javaClassName.indexOf('.') >= 0) {
	  endIndex = 1;  // if name contains package, don't check imports
      }
      for (int i=0; i<endIndex; ++i) {
	  try {
	      return Class.forName(imported[i]+javaClassName);
	  } catch (ClassNotFoundException cnf1) { }
	  catch (IllegalArgumentException iae) {
	      break;
	  }
      }
      throw new TypeException();
  }


}
